# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import datetime
import os
import shutil
import socket
import ssl

import tornado.autoreload
import tornado.locale
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.x509 import ExtensionNotFound
from mako.lookup import TemplateLookup
from tornado.httpserver import HTTPServer
from tornado.ioloop import PeriodicCallback
from tornado.web import Application, RedirectHandler, StaticFileHandler

import sickrage
from sickrage.core.helpers import create_https_certificates
from sickrage.core.webserver.handlers.account import AccountLinkHandler, AccountUnlinkHandler, AccountIsLinkedHandler
from sickrage.core.webserver.handlers.announcements import AnnouncementsHandler, MarkAnnouncementSeenHandler, AnnouncementCountHandler
from sickrage.core.webserver.handlers.api import ApiSwaggerDotJsonHandler, ApiPingHandler, ApiProfileHandler
from sickrage.core.webserver.handlers.api.v1 import ApiV1Handler
from sickrage.core.webserver.handlers.api.v2 import ApiV2RetrieveSeriesMetadataHandler
from sickrage.core.webserver.handlers.api.v2.config import ApiV2ConfigHandler
from sickrage.core.webserver.handlers.api.v2.file_browser import ApiV2FileBrowserHandler
from sickrage.core.webserver.handlers.api.v2.history import ApiV2HistoryHandler
from sickrage.core.webserver.handlers.api.v2.postprocess import Apiv2PostProcessHandler
from sickrage.core.webserver.handlers.api.v2.schedule import ApiV2ScheduleHandler
from sickrage.core.webserver.handlers.api.v2.series import ApiV2SeriesHandler, ApiV2SeriesEpisodesHandler, ApiV2SeriesImagesHandler, ApiV2SeriesImdbInfoHandler, \
    ApiV2SeriesBlacklistHandler, ApiV2SeriesWhitelistHandler, ApiV2SeriesRefreshHandler, ApiV2SeriesUpdateHandler, ApiV2SeriesEpisodesRenameHandler, \
    ApiV2SeriesEpisodesManualSearchHandler
from sickrage.core.webserver.handlers.api.v2.series_provider import ApiV2SeriesProvidersHandler, ApiV2SeriesProvidersSearchHandler, \
    ApiV2SeriesProvidersLanguagesHandler
from sickrage.core.webserver.handlers.calendar import CalendarHandler
from sickrage.core.webserver.handlers.changelog import ChangelogHandler
from sickrage.core.webserver.handlers.config import ConfigWebHandler, ConfigResetHandler
from sickrage.core.webserver.handlers.config.anime import ConfigAnimeHandler, ConfigSaveAnimeHandler
from sickrage.core.webserver.handlers.config.backup_restore import ConfigBackupRestoreHandler, ConfigBackupHandler, \
    ConfigRestoreHandler, SaveBackupRestoreHandler
from sickrage.core.webserver.handlers.config.general import GenerateApiKeyHandler, SaveRootDirsHandler, \
    SaveAddShowDefaultsHandler, SaveGeneralHandler, ConfigGeneralHandler
from sickrage.core.webserver.handlers.config.notifications import ConfigNotificationsHandler, SaveNotificationsHandler
from sickrage.core.webserver.handlers.config.postprocessing import ConfigPostProcessingHandler, \
    SavePostProcessingHandler, TestNamingHandler, IsRarSupportedHandler, IsNamingPatternValidHandler
from sickrage.core.webserver.handlers.config.providers import ConfigProvidersHandler, CanAddNewznabProviderHandler, \
    CanAddTorrentRssProviderHandler, GetNewznabCategoriesHandler, SaveProvidersHandler
from sickrage.core.webserver.handlers.config.quality_settings import ConfigQualitySettingsHandler, SaveQualitiesHandler
from sickrage.core.webserver.handlers.config.search import ConfigSearchHandler, SaveSearchHandler
from sickrage.core.webserver.handlers.config.subtitles import ConfigSubtitlesHandler, ConfigSubtitleGetCodeHandler, \
    ConfigSubtitlesWantedLanguagesHandler, SaveSubtitlesHandler
from sickrage.core.webserver.handlers.history import HistoryHandler, HistoryTrimHandler, HistoryClearHandler
from sickrage.core.webserver.handlers.home import HomeHandler, IsAliveHandler, TestSABnzbdHandler, TestTorrentHandler, \
    TestFreeMobileHandler, TestTelegramHandler, TestJoinHandler, TestGrowlHandler, TestProwlHandler, TestBoxcar2Handler, \
    TestPushoverHandler, FetchReleasegroupsHandler, RetryEpisodeHandler, TwitterStep1Handler, TwitterStep2Handler, \
    TestTwitterHandler, TestTwilioHandler, TestSlackHandler, TestDiscordHandler, TestKODIHandler, TestPMCHandler, \
    TestPMSHandler, TestLibnotifyHandler, TestEMBYHandler, TestNMJHandler, SettingsNMJHandler, TestNMJv2Handler, \
    SettingsNMJv2Handler, GetTraktTokenHandler, TestTraktHandler, LoadShowNotifyListsHandler, SaveShowNotifyListHandler, \
    TestEmailHandler, TestNMAHandler, TestPushalotHandler, TestPushbulletHandler, GetPushbulletDevicesHandler, \
    ShutdownHandler, RestartHandler, UpdateCheckHandler, UpdateHandler, VerifyPathHandler, \
    InstallRequirementsHandler, BranchCheckoutHandler, DisplayShowHandler, TogglePauseHandler, \
    DeleteShowHandler, RefreshShowHandler, UpdateShowHandler, SubtitleShowHandler, UpdateKODIHandler, UpdatePLEXHandler, \
    UpdateEMBYHandler, SyncTraktHandler, DeleteEpisodeHandler, TestRenameHandler, DoRenameHandler, \
    SearchEpisodeHandler, GetManualSearchStatusHandler, SearchEpisodeSubtitlesHandler, \
    SetSceneNumberingHandler, ProviderStatusHandler, ServerStatusHandler, ShowProgressHandler, TestSynologyDSMHandler, TestAlexaHandler
from sickrage.core.webserver.handlers.home.add_shows import HomeAddShowsHandler, SearchSeriesProviderForShowNameHandler, \
    MassAddTableHandler, NewShowHandler, TraktShowsHandler, PopularShowsHandler, AddShowToBlacklistHandler, \
    ExistingShowsHandler, AddShowByIDHandler, AddNewShowHandler, AddExistingShowsHandler
from sickrage.core.webserver.handlers.home.postprocess import HomePostProcessHandler, HomeProcessEpisodeHandler
from sickrage.core.webserver.handlers.login import LoginHandler
from sickrage.core.webserver.handlers.logout import LogoutHandler
from sickrage.core.webserver.handlers.logs import LogsHandler, LogsClearAllHanlder, LogsViewHandler, \
    LogsClearErrorsHanlder, LogsClearWarningsHanlder, ErrorCountHandler, WarningCountHandler
from sickrage.core.webserver.handlers.manage import ManageHandler, ShowEpisodeStatusesHandler, EpisodeStatusesHandler, \
    ChangeEpisodeStatusesHandler, ShowSubtitleMissedHandler, SubtitleMissedHandler, DownloadSubtitleMissedHandler, \
    BacklogShowHandler, BacklogOverviewHandler, MassEditHandler, MassUpdateHandler, FailedDownloadsHandler, EditShowHandler, SetEpisodeStatusHandler
from sickrage.core.webserver.handlers.manage.queues import ManageQueuesHandler, ForceBacklogSearchHandler, \
    ForceFindPropersHandler, PauseDailySearcherHandler, PauseBacklogSearcherHandler, PausePostProcessorHandler, \
    ForceDailySearchHandler
from sickrage.core.webserver.handlers.not_found import NotFoundHandler
from sickrage.core.webserver.handlers.root import RobotsDotTxtHandler, MessagesDotPoHandler, \
    APIBulderHandler, SetHomeLayoutHandler, SetPosterSortByHandler, SetPosterSortDirHandler, \
    ToggleDisplayShowSpecialsHandler, SetScheduleLayoutHandler, ToggleScheduleDisplayPausedHandler, \
    SetScheduleSortHandler, ScheduleHandler, QuicksearchDotJsonHandler, SetHistoryLayoutHandler, ForceSchedulerJobHandler
from sickrage.core.webserver.handlers.web_file_browser import WebFileBrowserHandler, WebFileBrowserCompleteHandler
from sickrage.core.websocket import WebSocketUIHandler


class StaticImageHandler(StaticFileHandler):
    def initialize(self, path, default_filename=None):
        super(StaticImageHandler, self).initialize(path, default_filename)

    def get(self, path, include_body=True):
        # image cache check
        self.root = (self.root, os.path.join(sickrage.app.cache_dir, 'images'))[
            os.path.exists(os.path.normpath(os.path.join(sickrage.app.cache_dir, 'images', path)))
        ]

        return super(StaticImageHandler, self).get(path, include_body)


class StaticNoCacheFileHandler(StaticFileHandler):
    def set_extra_headers(self, path):
        self.set_header('Cache-Control', 'max-age=0,no-cache,no-store')


class WebServer(object):
    def __init__(self):
        super(WebServer, self).__init__()
        self.name = "TORNADO"
        self.daemon = True
        self.started = False
        self.handlers = {}
        self.video_root = None
        self.api_v1_root = None
        self.api_v2_root = None
        self.app = None
        self.server = None

    def start(self):
        self.started = True

        # watch websocket message queue
        PeriodicCallback(self.check_web_socket_queue, 100).start()

        # load languages
        tornado.locale.load_gettext_translations(sickrage.LOCALE_DIR, 'messages')

        # Check configured web port is correct
        if sickrage.app.config.general.web_port < 21 or sickrage.app.config.general.web_port > 65535:
            sickrage.app.config.general.web_port = 8081

        # clear mako cache folder
        mako_cache = os.path.join(sickrage.app.cache_dir, 'mako')
        if os.path.isdir(mako_cache):
            shutil.rmtree(mako_cache, ignore_errors=True)

        # video root
        if sickrage.app.config.general.root_dirs:
            root_dirs = sickrage.app.config.general.root_dirs.split('|')
            self.video_root = root_dirs[int(root_dirs[0]) + 1]

        # web root
        if sickrage.app.config.general.web_root:
            sickrage.app.config.general.web_root = sickrage.app.config.general.web_root = ('/' + sickrage.app.config.general.web_root.lstrip('/').strip('/'))

        # api root
        self.api_v1_root = fr'{sickrage.app.config.general.web_root}/api/(?:v1/)?({sickrage.app.config.general.api_v1_key})'
        self.api_v2_root = fr'{sickrage.app.config.general.web_root}/api/v2'

        # tornado SSL setup
        if sickrage.app.config.general.enable_https:
            if not self.load_ssl_certificate():
                sickrage.app.log.info("Unable to retrieve CERT/KEY files from SiCKRAGE API, disabling HTTPS")
                sickrage.app.config.general.enable_https = False

        # Load templates
        mako_lookup = TemplateLookup(
            directories=[sickrage.app.gui_views_dir],
            module_directory=os.path.join(sickrage.app.cache_dir, 'mako'),
            filesystem_checks=True,
            strict_undefined=True,
            input_encoding='utf-8',
            output_encoding='utf-8',
            encoding_errors='replace'
        )

        templates = {}
        for root, dirs, files in os.walk(sickrage.app.gui_views_dir):
            path = root.split(os.sep)

            for x in sickrage.app.gui_views_dir.split(os.sep):
                if x in path:
                    del path[path.index(x)]

            for file in files:
                filename = '{}/{}'.format('/'.join(path), file).lstrip('/')
                templates[filename] = mako_lookup.get_template(filename)

        # Websocket handler
        self.handlers['websocket_handlers'] = [
            (fr'{sickrage.app.config.general.web_root}/ws/ui', WebSocketUIHandler)
        ]

        # API v1 Handlers
        self.handlers['api_v1_handlers'] = [
            # api
            (fr'{self.api_v1_root}(/?.*)', ApiV1Handler),

            # api builder
            (fr'{sickrage.app.config.general.web_root}/api/builder', RedirectHandler,
             {"url": sickrage.app.config.general.web_root + '/apibuilder/'}),
        ]

        # API v2 Handlers
        self.handlers['api_v2_handlers'] = [
            (fr'{self.api_v2_root}/ping', ApiPingHandler),
            (fr'{self.api_v2_root}/profile', ApiProfileHandler),
            (fr'{self.api_v2_root}/swagger.json', ApiSwaggerDotJsonHandler, {'api_handlers': 'api_v2_handlers', 'api_version': '2.0.0'}),
            (fr'{self.api_v2_root}/config', ApiV2ConfigHandler),
            (fr'{self.api_v2_root}/file-browser', ApiV2FileBrowserHandler),
            (fr'{self.api_v2_root}/postprocess', Apiv2PostProcessHandler),
            (fr'{self.api_v2_root}/retrieve-series-metadata', ApiV2RetrieveSeriesMetadataHandler),
            (fr'{self.api_v2_root}/schedule', ApiV2ScheduleHandler),
            (fr'{self.api_v2_root}/history', ApiV2HistoryHandler),
            (fr'{self.api_v2_root}/series-providers', ApiV2SeriesProvidersHandler),
            (fr'{self.api_v2_root}/series-providers/([a-z]+)/search', ApiV2SeriesProvidersSearchHandler),
            (fr'{self.api_v2_root}/series-providers/([a-z]+)/languages', ApiV2SeriesProvidersLanguagesHandler),
            (fr'{self.api_v2_root}/series', ApiV2SeriesHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)', ApiV2SeriesHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/episodes', ApiV2SeriesEpisodesHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/episodes/rename', ApiV2SeriesEpisodesRenameHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/episodes/(s\d+e\d+)/search', ApiV2SeriesEpisodesManualSearchHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/images', ApiV2SeriesImagesHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/imdb-info', ApiV2SeriesImdbInfoHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/blacklist', ApiV2SeriesBlacklistHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/whitelist', ApiV2SeriesWhitelistHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/refresh', ApiV2SeriesRefreshHandler),
            (fr'{self.api_v2_root}/series/(\d+[-][a-z]+)/update', ApiV2SeriesUpdateHandler)
        ]

        # New UI Static File Handlers
        self.handlers['new_ui_static_file_handlers'] = [
            # media
            (fr'{sickrage.app.config.general.web_root}/app/static/media/(.*)', StaticImageHandler,
             {"path": os.path.join(sickrage.app.gui_app_dir, 'static', 'media')}),

            # css
            (fr'{sickrage.app.config.general.web_root}/app/static/css/(.*)', StaticNoCacheFileHandler,
             {"path": os.path.join(sickrage.app.gui_app_dir, 'static', 'css')}),

            # js
            (fr'{sickrage.app.config.general.web_root}/app/static/js/(.*)', StaticNoCacheFileHandler,
             {"path": os.path.join(sickrage.app.gui_app_dir, 'static', 'js')}),

            # base
            (fr"{sickrage.app.config.general.web_root}/app/(.*)", tornado.web.StaticFileHandler,
             {"path": sickrage.app.gui_app_dir, "default_filename": "index.html"})
        ]

        # Static File Handlers
        self.handlers['static_file_handlers'] = [
            # redirect to home
            (fr"({sickrage.app.config.general.web_root})(/?)", RedirectHandler,
             {"url": f"{sickrage.app.config.general.web_root}/home"}),

            # login
            (fr'{sickrage.app.config.general.web_root}/login(/?)', LoginHandler),

            # logout
            (fr'{sickrage.app.config.general.web_root}/logout(/?)', LogoutHandler),

            # favicon
            (fr'{sickrage.app.config.general.web_root}/(favicon\.ico)', StaticNoCacheFileHandler,
             {"path": os.path.join(sickrage.app.gui_static_dir, 'images/favicon.ico')}),

            # images
            (fr'{sickrage.app.config.general.web_root}/images/(.*)', StaticImageHandler,
             {"path": os.path.join(sickrage.app.gui_static_dir, 'images')}),

            # css
            (fr'{sickrage.app.config.general.web_root}/css/(.*)', StaticNoCacheFileHandler,
             {"path": os.path.join(sickrage.app.gui_static_dir, 'css')}),

            # scss
            (fr'{sickrage.app.config.general.web_root}/scss/(.*)', StaticNoCacheFileHandler,
             {"path": os.path.join(sickrage.app.gui_static_dir, 'scss')}),

            # fonts
            (fr'{sickrage.app.config.general.web_root}/fonts/(.*)', StaticNoCacheFileHandler,
             {"path": os.path.join(sickrage.app.gui_static_dir, 'fonts')}),

            # javascript
            (fr'{sickrage.app.config.general.web_root}/js/(.*)', StaticNoCacheFileHandler,
             {"path": os.path.join(sickrage.app.gui_static_dir, 'js')}),

            # videos
            (fr'{sickrage.app.config.general.web_root}/videos/(.*)', StaticNoCacheFileHandler,
             {"path": self.video_root}),
        ]

        # Handlers
        self.handlers['web_handlers'] = [
            (fr'{sickrage.app.config.general.web_root}/robots.txt', RobotsDotTxtHandler),
            (fr'{sickrage.app.config.general.web_root}/messages.po', MessagesDotPoHandler),
            (fr'{sickrage.app.config.general.web_root}/quicksearch.json', QuicksearchDotJsonHandler),
            (fr'{sickrage.app.config.general.web_root}/apibuilder(/?)', APIBulderHandler),
            (fr'{sickrage.app.config.general.web_root}/setHomeLayout(/?)', SetHomeLayoutHandler),
            (fr'{sickrage.app.config.general.web_root}/setPosterSortBy(/?)', SetPosterSortByHandler),
            (fr'{sickrage.app.config.general.web_root}/setPosterSortDir(/?)', SetPosterSortDirHandler),
            (fr'{sickrage.app.config.general.web_root}/setHistoryLayout(/?)', SetHistoryLayoutHandler),
            (fr'{sickrage.app.config.general.web_root}/toggleDisplayShowSpecials(/?)', ToggleDisplayShowSpecialsHandler),
            (fr'{sickrage.app.config.general.web_root}/toggleScheduleDisplayPaused(/?)', ToggleScheduleDisplayPausedHandler),
            (fr'{sickrage.app.config.general.web_root}/setScheduleSort(/?)', SetScheduleSortHandler),
            (fr'{sickrage.app.config.general.web_root}/forceSchedulerJob(/?)', ForceSchedulerJobHandler),
            (fr'{sickrage.app.config.general.web_root}/announcements(/?)', AnnouncementsHandler),
            (fr'{sickrage.app.config.general.web_root}/announcements/announcementCount(/?)', AnnouncementCountHandler),
            (fr'{sickrage.app.config.general.web_root}/announcements/mark-seen(/?)', MarkAnnouncementSeenHandler),
            (fr'{sickrage.app.config.general.web_root}/schedule(/?)', ScheduleHandler),
            (fr'{sickrage.app.config.general.web_root}/setScheduleLayout(/?)', SetScheduleLayoutHandler),
            (fr'{sickrage.app.config.general.web_root}/calendar(/?)', CalendarHandler),
            (fr'{sickrage.app.config.general.web_root}/changelog(/?)', ChangelogHandler),
            (fr'{sickrage.app.config.general.web_root}/account/link(/?)', AccountLinkHandler),
            (fr'{sickrage.app.config.general.web_root}/account/unlink(/?)', AccountUnlinkHandler),
            (fr'{sickrage.app.config.general.web_root}/account/is-linked(/?)', AccountIsLinkedHandler),
            (fr'{sickrage.app.config.general.web_root}/history(/?)', HistoryHandler),
            (fr'{sickrage.app.config.general.web_root}/history/clear(/?)', HistoryClearHandler),
            (fr'{sickrage.app.config.general.web_root}/history/trim(/?)', HistoryTrimHandler),
            (fr'{sickrage.app.config.general.web_root}/logs(/?)', LogsHandler),
            (fr'{sickrage.app.config.general.web_root}/logs/errorCount(/?)', ErrorCountHandler),
            (fr'{sickrage.app.config.general.web_root}/logs/warningCount(/?)', WarningCountHandler),
            (fr'{sickrage.app.config.general.web_root}/logs/view(/?)', LogsViewHandler),
            (fr'{sickrage.app.config.general.web_root}/logs/clearAll(/?)', LogsClearAllHanlder),
            (fr'{sickrage.app.config.general.web_root}/logs/clearWarnings(/?)', LogsClearWarningsHanlder),
            (fr'{sickrage.app.config.general.web_root}/logs/clearErrors(/?)', LogsClearErrorsHanlder),
            (fr'{sickrage.app.config.general.web_root}/browser(/?)', WebFileBrowserHandler),
            (fr'{sickrage.app.config.general.web_root}/browser/complete(/?)', WebFileBrowserCompleteHandler),
            (fr'{sickrage.app.config.general.web_root}/home(/?)', HomeHandler),
            (fr'{sickrage.app.config.general.web_root}/home/showProgress(/?)', ShowProgressHandler),
            (fr'{sickrage.app.config.general.web_root}/home/is-alive(/?)', IsAliveHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testSABnzbd(/?)', TestSABnzbdHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testSynologyDSM(/?)', TestSynologyDSMHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testTorrent(/?)', TestTorrentHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testFreeMobile(/?)', TestFreeMobileHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testTelegram(/?)', TestTelegramHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testJoin(/?)', TestJoinHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testGrowl(/?)', TestGrowlHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testProwl(/?)', TestProwlHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testBoxcar2(/?)', TestBoxcar2Handler),
            (fr'{sickrage.app.config.general.web_root}/home/testPushover(/?)', TestPushoverHandler),
            (fr'{sickrage.app.config.general.web_root}/home/twitterStep1(/?)', TwitterStep1Handler),
            (fr'{sickrage.app.config.general.web_root}/home/twitterStep2(/?)', TwitterStep2Handler),
            (fr'{sickrage.app.config.general.web_root}/home/testTwitter(/?)', TestTwitterHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testTwilio(/?)', TestTwilioHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testSlack(/?)', TestSlackHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testAlexa(/?)', TestAlexaHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testDiscord(/?)', TestDiscordHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testKODI(/?)', TestKODIHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testPMC(/?)', TestPMCHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testPMS(/?)', TestPMSHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testLibnotify(/?)', TestLibnotifyHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testEMBY(/?)', TestEMBYHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testNMJ(/?)', TestNMJHandler),
            (fr'{sickrage.app.config.general.web_root}/home/settingsNMJ(/?)', SettingsNMJHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testNMJv2(/?)', TestNMJv2Handler),
            (fr'{sickrage.app.config.general.web_root}/home/settingsNMJv2(/?)', SettingsNMJv2Handler),
            (fr'{sickrage.app.config.general.web_root}/home/getTraktToken(/?)', GetTraktTokenHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testTrakt(/?)', TestTraktHandler),
            (fr'{sickrage.app.config.general.web_root}/home/loadShowNotifyLists(/?)', LoadShowNotifyListsHandler),
            (fr'{sickrage.app.config.general.web_root}/home/saveShowNotifyList(/?)', SaveShowNotifyListHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testEmail(/?)', TestEmailHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testNMA(/?)', TestNMAHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testPushalot(/?)', TestPushalotHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testPushbullet(/?)', TestPushbulletHandler),
            (fr'{sickrage.app.config.general.web_root}/home/getPushbulletDevices(/?)', GetPushbulletDevicesHandler),
            (fr'{sickrage.app.config.general.web_root}/home/serverStatus(/?)', ServerStatusHandler),
            (fr'{sickrage.app.config.general.web_root}/home/providerStatus(/?)', ProviderStatusHandler),
            (fr'{sickrage.app.config.general.web_root}/home/shutdown(/?)', ShutdownHandler),
            (fr'{sickrage.app.config.general.web_root}/home/restart(/?)', RestartHandler),
            (fr'{sickrage.app.config.general.web_root}/home/updateCheck(/?)', UpdateCheckHandler),
            (fr'{sickrage.app.config.general.web_root}/home/update(/?)', UpdateHandler),
            (fr'{sickrage.app.config.general.web_root}/home/verifyPath(/?)', VerifyPathHandler),
            (fr'{sickrage.app.config.general.web_root}/home/installRequirements(/?)', InstallRequirementsHandler),
            (fr'{sickrage.app.config.general.web_root}/home/branchCheckout(/?)', BranchCheckoutHandler),
            (fr'{sickrage.app.config.general.web_root}/home/displayShow(/?)', DisplayShowHandler),
            (fr'{sickrage.app.config.general.web_root}/home/togglePause(/?)', TogglePauseHandler),
            (fr'{sickrage.app.config.general.web_root}/home/deleteShow', DeleteShowHandler),
            (fr'{sickrage.app.config.general.web_root}/home/refreshShow(/?)', RefreshShowHandler),
            (fr'{sickrage.app.config.general.web_root}/home/updateShow(/?)', UpdateShowHandler),
            (fr'{sickrage.app.config.general.web_root}/home/subtitleShow(/?)', SubtitleShowHandler),
            (fr'{sickrage.app.config.general.web_root}/home/updateKODI(/?)', UpdateKODIHandler),
            (fr'{sickrage.app.config.general.web_root}/home/updatePLEX(/?)', UpdatePLEXHandler),
            (fr'{sickrage.app.config.general.web_root}/home/updateEMBY(/?)', UpdateEMBYHandler),
            (fr'{sickrage.app.config.general.web_root}/home/syncTrakt(/?)', SyncTraktHandler),
            (fr'{sickrage.app.config.general.web_root}/home/deleteEpisode(/?)', DeleteEpisodeHandler),
            (fr'{sickrage.app.config.general.web_root}/home/testRename(/?)', TestRenameHandler),
            (fr'{sickrage.app.config.general.web_root}/home/doRename(/?)', DoRenameHandler),
            (fr'{sickrage.app.config.general.web_root}/home/searchEpisode(/?)', SearchEpisodeHandler),
            (fr'{sickrage.app.config.general.web_root}/home/getManualSearchStatus(/?)', GetManualSearchStatusHandler),
            (fr'{sickrage.app.config.general.web_root}/home/searchEpisodeSubtitles(/?)', SearchEpisodeSubtitlesHandler),
            (fr'{sickrage.app.config.general.web_root}/home/setSceneNumbering(/?)', SetSceneNumberingHandler),
            (fr'{sickrage.app.config.general.web_root}/home/retryEpisode(/?)', RetryEpisodeHandler),
            (fr'{sickrage.app.config.general.web_root}/home/fetch_releasegroups(/?)', FetchReleasegroupsHandler),
            (fr'{sickrage.app.config.general.web_root}/home/postprocess(/?)', HomePostProcessHandler),
            (fr'{sickrage.app.config.general.web_root}/home/postprocess/processEpisode(/?)', HomeProcessEpisodeHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows(/?)', HomeAddShowsHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/searchSeriesProviderForShowName(/?)', SearchSeriesProviderForShowNameHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/massAddTable(/?)', MassAddTableHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/newShow(/?)', NewShowHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/traktShows(/?)', TraktShowsHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/popularShows(/?)', PopularShowsHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/addShowToBlacklist(/?)', AddShowToBlacklistHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/existingShows(/?)', ExistingShowsHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/addShowByID(/?)', AddShowByIDHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/addNewShow(/?)', AddNewShowHandler),
            (fr'{sickrage.app.config.general.web_root}/home/addShows/addExistingShows(/?)', AddExistingShowsHandler),
            (fr'{sickrage.app.config.general.web_root}/manage(/?)', ManageHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/editShow(/?)', EditShowHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/showEpisodeStatuses(/?)', ShowEpisodeStatusesHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/episodeStatuses(/?)', EpisodeStatusesHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/changeEpisodeStatuses(/?)', ChangeEpisodeStatusesHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/setEpisodeStatus(/?)', SetEpisodeStatusHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/showSubtitleMissed(/?)', ShowSubtitleMissedHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/subtitleMissed(/?)', SubtitleMissedHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/downloadSubtitleMissed(/?)', DownloadSubtitleMissedHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/backlogShow(/?)', BacklogShowHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/backlogOverview(/?)', BacklogOverviewHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/massEdit(/?)', MassEditHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/massUpdate(/?)', MassUpdateHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/failedDownloads(/?)', FailedDownloadsHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/manageQueues(/?)', ManageQueuesHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/manageQueues/forceBacklogSearch(/?)', ForceBacklogSearchHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/manageQueues/forceDailySearch(/?)', ForceDailySearchHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/manageQueues/forceFindPropers(/?)', ForceFindPropersHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/manageQueues/pauseDailySearcher(/?)', PauseDailySearcherHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/manageQueues/pauseBacklogSearcher(/?)', PauseBacklogSearcherHandler),
            (fr'{sickrage.app.config.general.web_root}/manage/manageQueues/pausePostProcessor(/?)', PausePostProcessorHandler),
            (fr'{sickrage.app.config.general.web_root}/config(/?)', ConfigWebHandler),
            (fr'{sickrage.app.config.general.web_root}/config/reset(/?)', ConfigResetHandler),
            (fr'{sickrage.app.config.general.web_root}/config/anime(/?)', ConfigAnimeHandler),
            (fr'{sickrage.app.config.general.web_root}/config/anime/saveAnime(/?)', ConfigSaveAnimeHandler),
            (fr'{sickrage.app.config.general.web_root}/config/backuprestore(/?)', ConfigBackupRestoreHandler),
            (fr'{sickrage.app.config.general.web_root}/config/backuprestore/backup(/?)', ConfigBackupHandler),
            (fr'{sickrage.app.config.general.web_root}/config/backuprestore/restore(/?)', ConfigRestoreHandler),
            (fr'{sickrage.app.config.general.web_root}/config/backuprestore/saveBackupRestore(/?)', SaveBackupRestoreHandler),
            (fr'{sickrage.app.config.general.web_root}/config/general(/?)', ConfigGeneralHandler),
            (fr'{sickrage.app.config.general.web_root}/config/general/generateApiKey(/?)', GenerateApiKeyHandler),
            (fr'{sickrage.app.config.general.web_root}/config/general/saveRootDirs(/?)', SaveRootDirsHandler),
            (fr'{sickrage.app.config.general.web_root}/config/general/saveAddShowDefaults(/?)', SaveAddShowDefaultsHandler),
            (fr'{sickrage.app.config.general.web_root}/config/general/saveGeneral(/?)', SaveGeneralHandler),
            (fr'{sickrage.app.config.general.web_root}/config/notifications(/?)', ConfigNotificationsHandler),
            (fr'{sickrage.app.config.general.web_root}/config/notifications/saveNotifications(/?)', SaveNotificationsHandler),
            (fr'{sickrage.app.config.general.web_root}/config/postProcessing(/?)', ConfigPostProcessingHandler),
            (fr'{sickrage.app.config.general.web_root}/config/postProcessing/savePostProcessing(/?)', SavePostProcessingHandler),
            (fr'{sickrage.app.config.general.web_root}/config/postProcessing/testNaming(/?)', TestNamingHandler),
            (fr'{sickrage.app.config.general.web_root}/config/postProcessing/isNamingValid(/?)', IsNamingPatternValidHandler),
            (fr'{sickrage.app.config.general.web_root}/config/postProcessing/isRarSupported(/?)', IsRarSupportedHandler),
            (fr'{sickrage.app.config.general.web_root}/config/providers(/?)', ConfigProvidersHandler),
            (fr'{sickrage.app.config.general.web_root}/config/providers/canAddNewznabProvider(/?)', CanAddNewznabProviderHandler),
            (fr'{sickrage.app.config.general.web_root}/config/providers/canAddTorrentRssProvider(/?)', CanAddTorrentRssProviderHandler),
            (fr'{sickrage.app.config.general.web_root}/config/providers/getNewznabCategories(/?)', GetNewznabCategoriesHandler),
            (fr'{sickrage.app.config.general.web_root}/config/providers/saveProviders(/?)', SaveProvidersHandler),
            (fr'{sickrage.app.config.general.web_root}/config/qualitySettings(/?)', ConfigQualitySettingsHandler),
            (fr'{sickrage.app.config.general.web_root}/config/qualitySettings/saveQualities(/?)', SaveQualitiesHandler),
            (fr'{sickrage.app.config.general.web_root}/config/search(/?)', ConfigSearchHandler),
            (fr'{sickrage.app.config.general.web_root}/config/search/saveSearch(/?)', SaveSearchHandler),
            (fr'{sickrage.app.config.general.web_root}/config/subtitles(/?)', ConfigSubtitlesHandler),
            (fr'{sickrage.app.config.general.web_root}/config/subtitles/get_code(/?)', ConfigSubtitleGetCodeHandler),
            (fr'{sickrage.app.config.general.web_root}/config/subtitles/wanted_languages(/?)', ConfigSubtitlesWantedLanguagesHandler),
            (fr'{sickrage.app.config.general.web_root}/config/subtitles/saveSubtitles(/?)', SaveSubtitlesHandler),
        ]

        # Initialize Tornado application
        self.app = Application(
            handlers=sum(self.handlers.values(), []),
            debug=True,
            autoreload=False,
            gzip=sickrage.app.config.general.web_use_gzip,
            cookie_secret=sickrage.app.config.general.web_cookie_secret,
            login_url='%s/login/' % sickrage.app.config.general.web_root,
            templates=templates,
            default_handler_class=NotFoundHandler
        )

        # HTTPS Cert/Key object
        ssl_ctx = None
        if sickrage.app.config.general.enable_https:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(sickrage.app.config.general.https_cert, sickrage.app.config.general.https_key)

        # Web Server
        self.server = HTTPServer(self.app, ssl_options=ssl_ctx, xheaders=sickrage.app.config.general.handle_reverse_proxy)

        try:
            self.server.listen(sickrage.app.config.general.web_port, sickrage.app.web_host)
        except socket.error as e:
            sickrage.app.log.warning(e.strerror)
            raise SystemExit

    def check_web_socket_queue(self):
        if not WebSocketUIHandler.message_queue.empty():
            message = WebSocketUIHandler.message_queue.get()
            WebSocketUIHandler.broadcast(message)

    def load_ssl_certificate(self, certificate=None, private_key=None):
        if certificate and private_key:
            with open(sickrage.app.config.general.https_cert, 'w') as cert_out:
                cert_out.write(certificate)

            with open(sickrage.app.config.general.https_key, 'w') as key_out:
                key_out.write(private_key)
        else:
            if os.path.exists(sickrage.app.config.general.https_key) and os.path.exists(sickrage.app.config.general.https_cert):
                if self.is_certificate_valid() and not self.certificate_needs_renewal():
                    return True

            resp = sickrage.app.api.server.get_server_certificate(sickrage.app.config.general.server_id)
            if not resp or 'certificate' not in resp or 'private_key' not in resp:
                if not create_https_certificates(sickrage.app.config.general.https_cert, sickrage.app.config.general.https_key):
                    return False

                if not os.path.exists(sickrage.app.config.general.https_cert) or not os.path.exists(sickrage.app.config.general.https_key):
                    return False

                return True

            with open(sickrage.app.config.general.https_cert, 'w') as cert_out:
                cert_out.write(resp['certificate'])

            with open(sickrage.app.config.general.https_key, 'w') as key_out:
                key_out.write(resp['private_key'])

        sickrage.app.log.info("Loaded SSL certificate successfully")

        if self.server:
            sickrage.app.shutdown(restart=True)

        return True

    def certificate_needs_renewal(self):
        if not os.path.exists(sickrage.app.config.general.https_cert):
            return

        with open(sickrage.app.config.general.https_cert, 'rb') as f:
            cert_pem = f.read()

        cert = x509.load_pem_x509_certificate(cert_pem, default_backend())
        not_valid_after = cert.not_valid_after

        return not_valid_after - datetime.datetime.utcnow() < (cert.not_valid_after - cert.not_valid_before) / 2

    def is_certificate_valid(self):
        if not os.path.exists(sickrage.app.config.general.https_cert):
            return

        with open(sickrage.app.config.general.https_cert, 'rb') as f:
            cert_pem = f.read()

        cert = x509.load_pem_x509_certificate(cert_pem, default_backend())
        issuer = cert.issuer.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0]
        subject = cert.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0]

        if 'ZeroSSL' not in issuer.value:
            return False

        if subject.value != f'{sickrage.app.config.general.server_id}.external.sickrage.direct':
            return False

        try:
            ext = cert.extensions.get_extension_for_class(x509.SubjectAlternativeName).value
            sans = ext.get_values_for_type(x509.DNSName)

            domains = [
                f'{sickrage.app.config.general.server_id}.external.sickrage.direct',
                f'{sickrage.app.config.general.server_id}.internal.sickrage.direct'
            ]

            for domain in sans:
                if domain not in domains:
                    return False
        except ExtensionNotFound:
            return False

        return True

    def shutdown(self):
        if self.started:
            self.started = False
            if self.server:
                self.server.close_all_connections()
                self.server.stop()



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


from urllib.parse import urlencode

import requests

import sickrage
from sickrage.core.websession import WebSession
from sickrage.notification_providers import NotificationProvider


class PushalotNotification(NotificationProvider):
    def __init__(self):
        super(PushalotNotification, self).__init__()
        self.name = 'pushalot'

    def test_notify(self, pushalot_authorizationtoken):
        return self._sendPushalot(pushalot_authorizationtoken, event="Test",
                                  message="Testing Pushalot settings from SiCKRAGE", force=True)

    def notify_snatch(self, ep_name):
        if sickrage.app.config.pushalot.notify_on_snatch:
            self._sendPushalot(pushalot_authorizationtoken=sickrage.app.config.pushalot.auth_token,
                               event=self.notifyStrings[self.NOTIFY_SNATCH],
                               message=ep_name)

    def notify_download(self, ep_name):
        if sickrage.app.config.pushalot.notify_on_download:
            self._sendPushalot(pushalot_authorizationtoken=sickrage.app.config.pushalot.auth_token,
                               event=self.notifyStrings[self.NOTIFY_DOWNLOAD],
                               message=ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if sickrage.app.config.pushalot.notify_on_subtitle_download:
            self._sendPushalot(pushalot_authorizationtoken=sickrage.app.config.pushalot.auth_token,
                               event=self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD],
                               message=ep_name + ": " + lang)

    def notify_version_update(self, new_version="??"):
        if sickrage.app.config.pushalot.enable:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._sendPushalot(pushalot_authorizationtoken=sickrage.app.config.pushalot.auth_token,
                               event=title,
                               message=update_text + new_version)

    def _sendPushalot(self, pushalot_authorizationtoken=None, event=None, message=None, force=False):
        if not sickrage.app.config.pushalot.enable and not force:
            return False

        sickrage.app.log.debug("Pushalot event: " + event)
        sickrage.app.log.debug("Pushalot message: " + message)
        sickrage.app.log.debug("Pushalot api: " + pushalot_authorizationtoken)

        data = {'AuthorizationToken': pushalot_authorizationtoken,
                'Title': event,
                'Body': message}

        try:
            WebSession().post("https://pushalot.com/api/sendmessage",
                              headers={'Content-type': "application/x-www-form-urlencoded"},
                              data=urlencode(data))
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 410:
                sickrage.app.log.warning("Pushalot auth failed: %s" % e.response.text)
                return False

            sickrage.app.log.error("Pushalot notification failed.")
            return False

        sickrage.app.log.debug("Pushalot notifications sent.")
        return True



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.



import cgi
import os

import sickrage
from sickrage.notification_providers import NotificationProvider


def diagnose():
    '''
    Check the environment for reasons libnotify isn't working.  Return a
    user-readable message indicating possible issues.
    '''
    try:
        # noinspection PyUnresolvedReferences
        from gi.repository import Notify  # @UnusedImport
    except ImportError:
        return ("<p>Error: gir-notify isn't installed. On Ubuntu/Debian, install the "
                "<a href=\"apt:gir1.2-notify-0.7\">gir1.2-notify-0.7</a> or "
                "<a href=\"apt:gir1.0-notify-0.4\">gir1.0-notify-0.4</a> package.")
    if 'DISPLAY' not in os.environ and 'DBUS_SESSION_BUS_ADDRESS' not in os.environ:
        return ("<p>Error: Environment variables DISPLAY and DBUS_SESSION_BUS_ADDRESS "
                "aren't set.  libnotify will only work when you run SiCKRAGE "
                "from a desktop login.")
    try:
        # noinspection PyUnresolvedReferences
        import dbus
    except ImportError:
        pass
    else:
        try:
            bus = dbus.SessionBus()
        except dbus.DBusException as e:
            return ("<p>Error: unable to connect to D-Bus session bus: <code>%s</code>."
                    "<p>Are you running SiCKRAGE in a desktop session?") % (cgi.escape(e),)
        try:
            bus.get_object('org.freedesktop.Notifications',
                           '/org/freedesktop/Notifications')
        except dbus.DBusException as e:
            return ("<p>Error: there doesn't seem to be a notification daemon available: <code>%s</code> "
                    "<p>Try installing notification-daemon or notify-osd.") % (cgi.escape(e),)
    return "<p>Error: Unable to send notification."


class LibnotifyNotification(NotificationProvider):
    def __init__(self):
        super(LibnotifyNotification, self).__init__()
        self.name = 'libnotify'
        self.Notify = None
        self.gobject = None

    def init_notify(self):
        if self.Notify is not None:
            return True
        try:
            # noinspection PyUnresolvedReferences
            from gi.repository import Notify
        except ImportError:
            sickrage.app.log.error(
                "Unable to import Notify from gi.repository. libnotify notifications won't work.")
            return False
        try:
            # noinspection PyUnresolvedReferences
            from gi.repository import GObject
        except ImportError:
            sickrage.app.log.error(
                "Unable to import GObject from gi.repository. We can't catch a GError in display.")
            return False
        if not Notify.init('SiCKRAGE'):
            sickrage.app.log.error("Initialization of Notify failed. libnotify notifications won't work.")
            return False
        self.Notify = Notify
        self.gobject = GObject
        return True

    def notify_snatch(self, ep_name):
        if sickrage.app.config.libnotify.notify_on_snatch:
            self._notify(self.notifyStrings[self.NOTIFY_SNATCH], ep_name)

    def notify_download(self, ep_name):
        if sickrage.app.config.libnotify.notify_on_download:
            self._notify(self.notifyStrings[self.NOTIFY_DOWNLOAD], ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if sickrage.app.config.libnotify.notify_on_subtitle_download:
            self._notify(self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD], ep_name + ": " + lang)

    def notify_version_update(self, new_version="??"):
        if sickrage.app.config.libnotify.enable:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._notify(title, update_text + new_version)

    def test_notify(self):
        return self._notify('Test notification', "This is a test notification from SiCKRAGE", force=True)

    def _notify(self, title, message, force=False):
        if not sickrage.app.config.libnotify.enable and not force:
            return False
        if not self.init_notify():
            return False

        # Can't make this a global constant because PROG_DIR isn't available
        # when the module is imported.
        icon_path = os.path.join(sickrage.app.gui_static_dir, 'images', 'favicon.png')

        # If the session bus can't be acquired here a bunch of warning messages
        # will be printed but the call to show() will still return True.
        # pynotify doesn't seem too keen on error handling.
        n = self.Notify.Notification.new(title, message, icon_path)
        try:
            return n.show()
        except self.gobject.GError:
            return False



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import sickrage
from sickrage.core.enums import SeriesProviderID
from sickrage.core.webserver.handlers.api.v2 import ApiV2BaseHandler


class ApiV2SeriesProvidersHandler(ApiV2BaseHandler):
    def get(self):
        return self.json_response([{'displayName': x.display_name, 'slug': x.slug} for x in SeriesProviderID])


class ApiV2SeriesProvidersSearchHandler(ApiV2BaseHandler):
    def get(self, series_provider_slug):
        search_term = self.get_argument('searchTerm', None)
        lang = self.get_argument('seriesProviderLanguage', None)

        series_provider_id = SeriesProviderID.by_slug(series_provider_slug)
        if not series_provider_id:
            return self._bad_request(error="Unable to identify a series provider using provided slug")

        sickrage.app.log.debug(f"Searching for show with term: {search_term} on series provider: {sickrage.app.series_providers[series_provider_id].name}")

        # search via series name
        results = sickrage.app.series_providers[series_provider_id].search(search_term, language=lang)
        if not results:
            return self._not_found(error=f"Unable to find the series using the search term: {search_term}")

        return self.json_response(results)


class ApiV2SeriesProvidersLanguagesHandler(ApiV2BaseHandler):
    def get(self, series_provider_slug):
        series_provider_id = SeriesProviderID.by_slug(series_provider_slug)
        if not series_provider_id:
            return self._not_found(error="Unable to identify a series provider using provided slug")

        return self.json_response(sickrage.app.series_providers[series_provider_id].languages())



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import re
from urllib.parse import urljoin

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import convert_size
from sickrage.search_providers import TorrentProvider


class TorrentDayProvider(TorrentProvider):
    def __init__(self):
        super(TorrentDayProvider, self).__init__("TorrentDay", 'https://www.torrentday.com', True)

        self._urls.update({
            'login': '{base_url}/torrents/'.format(**self._urls),
            'search': '{base_url}/t.json'.format(**self._urls),
            'download': '{base_url}/download.php/'.format(**self._urls)
        })

        # custom settings
        self.custom_settings = {
            'username': '',
            'password': '',
            'freeleech': False,
            'minseed': 0,
            'minleech': 0
        }

        self.enable_cookies = True
        self.required_cookies = ('uid', 'pass')

        # TV/480p - 24
        # TV/Bluray - 32
        # TV/DVD-R - 31
        # TV/DVD-Rip - 33
        # TV/Mobile - 46
        # TV/Packs - 14
        # TV/SD/x264 - 26
        # TV/x264 - 7
        # TV/x265 - 34
        # TV/XviD - 2

        self.categories = {
            'Season': {'14': 1},
            'Episode': {'2': 1, '26': 1, '7': 1, '24': 1, '34': 1},
            'RSS': {'2': 1, '26': 1, '7': 1, '24': 1, '34': 1, '14': 1}
        }

        self.cache = TVCache(self)

    def login(self):
        return self.cookie_login('log in')

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        if not self.login():
            return results

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: %s" % mode)
            for search_string in search_strings[mode]:

                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: %s " % search_string)

                search_string = '+'.join(search_string.split())
                search_params = dict({'q': search_string}, **self.categories[mode])

                resp = self.session.get(self.urls['search'], params=search_params)
                if not resp or not resp.content:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                try:
                    data = resp.json()
                except ValueError:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(data, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        for item in data:
            try:
                # Check if this is a freeleech torrent and if we've configured to only allow freeleech.
                if self.custom_settings['freeleech'] and item.get('download-multiplier') != 0:
                    continue

                title = re.sub(r'\[.*\=.*\].*\[/.*\]', '', item['name']) if item['name'] else None
                download_url = urljoin(self.urls['download'], '{}/{}.torrent'.format(
                    item['t'], item['name']
                )) if item['t'] and item['name'] else None
                if not all([title, download_url]):
                    continue

                seeders = int(item['seeders'])
                leechers = int(item['leechers'])

                torrent_size = item['size']
                size = convert_size(torrent_size, -1)

                results += [{
                    'title': title,
                    'link': download_url,
                    'size': size,
                    'seeders': seeders,
                    'leechers': leechers
                }]

                if mode != 'RSS':
                    sickrage.app.log.debug("Found result: {}".format(title))
            except Exception:
                sickrage.app.log.error("Failed parsing provider.")

        return results



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import datetime
import os
import time
import traceback
from enum import Enum

import sickrage
from sickrage.core.common import EpisodeStatus
from sickrage.core.exceptions import CantRefreshShowException, CantRemoveShowException, CantUpdateShowException, EpisodeDeletedException, \
    MultipleShowObjectsException
from sickrage.core.queues import Queue, Task, TaskPriority
from sickrage.core.scene_numbering import xem_refresh
from sickrage.core.traktapi import TraktAPI
from sickrage.core.tv.show import TVShow
from sickrage.core.tv.show.helpers import find_show
from sickrage.core.websocket import WebSocketMessage
from sickrage.series_providers.exceptions import SeriesProviderAttributeNotFound, SeriesProviderException


class ShowQueue(Queue):
    def __init__(self):
        Queue.__init__(self, "SHOWQUEUE")

    @property
    def loading_show_list(self):
        return [x.series_id for x in self.tasks.copy().values() if x.is_loading]

    def _is_in_queue(self, series_id, actions):
        for task in self.tasks.copy().values():
            if task.is_queued() and task.series_id == series_id and task.action in actions:
                return True

        return False

    def _is_being(self, series_id, actions):
        for task in self.tasks.copy().values():
            if task.is_started() and task.series_id == series_id and task.action in actions:
                return True

        return False

    def is_queued_to_remove(self, series_id):
        return self._is_in_queue(series_id, [ShowTaskActions.REMOVE])

    def is_queued_to_add(self, series_id):
        return self._is_in_queue(series_id, [ShowTaskActions.ADD])

    def is_queued_to_update(self, series_id):
        return self._is_in_queue(series_id, [ShowTaskActions.UPDATE, ShowTaskActions.FORCEUPDATE])

    def is_queued_to_refresh(self, series_id):
        return self._is_in_queue(series_id, [ShowTaskActions.REFRESH])

    def is_queued_to_rename(self, series_id):
        return self._is_in_queue(series_id, [ShowTaskActions.RENAME])

    def is_queued_to_subtitle(self, series_id):
        return self._is_in_queue(series_id, [ShowTaskActions.SUBTITLE])

    def is_being_removed(self, series_id):
        return self._is_being(series_id, [ShowTaskActions.REMOVE])

    def is_being_added(self, series_id):
        return self._is_being(series_id, [ShowTaskActions.ADD])

    def is_being_updated(self, series_id):
        return self._is_being(series_id, [ShowTaskActions.UPDATE, ShowTaskActions.FORCEUPDATE])

    def is_being_refreshed(self, series_id):
        return self._is_being(series_id, [ShowTaskActions.REFRESH])

    def is_being_renamed(self, series_id):
        return self._is_being(series_id, [ShowTaskActions.RENAME])

    def is_being_subtitled(self, series_id):
        return self._is_being(series_id, [ShowTaskActions.SUBTITLE])

    def update_show(self, series_id, series_provider_id, force=False):
        show_obj = find_show(series_id, series_provider_id)

        if self.is_being_added(series_id):
            raise CantUpdateShowException("{} is still being added, please wait until it is finished before trying to update.".format(show_obj.name))

        if self.is_being_updated(series_id):
            raise CantUpdateShowException("{} is already being updated, can't update again until it's done.".format(show_obj.name))

        task_id = self.put(ShowTaskUpdate(series_id, series_provider_id, force))

        self.put(ShowTaskRefresh(series_id, series_provider_id, force=force), depend=[task_id])

    def refresh_show(self, series_id, series_provider_id, force=False):
        show_obj = find_show(series_id, series_provider_id)

        if (self.is_being_refreshed(series_id) or self.is_being_refreshed(series_id)) and not force:
            raise CantRefreshShowException("This show is already being refreshed or queued to be refreshed, skipping this request.")

        if show_obj.paused and not force:
            sickrage.app.log.debug('Skipping show [{}] because it is paused.'.format(show_obj.name))
            return

        sickrage.app.log.debug("Queueing show refresh for {}".format(show_obj.name))

        self.put(ShowTaskRefresh(series_id, series_provider_id, force=force))

    def rename_show_episodes(self, series_id, series_provider_id):
        self.put(ShowTaskRename(series_id, series_provider_id))

    def download_subtitles(self, series_id, series_provider_id):
        self.put(ShowTaskSubtitle(series_id, series_provider_id))

    def add_show(self, series_provider_id, series_id, showDir, default_status=None, quality=None, flatten_folders=None, lang=None, subtitles=None,
                 sub_use_sr_metadata=None, anime=None, search_format=None, dvd_order=None, paused=None, blacklist=None, whitelist=None,
                 default_status_after=None, scene=None, skip_downloaded=None):

        if lang is None:
            lang = sickrage.app.config.general.series_provider_default_language

        self.put(ShowTaskAdd(series_provider_id=series_provider_id,
                             series_id=series_id,
                             show_dir=showDir,
                             default_status=default_status,
                             quality=quality,
                             flatten_folders=not flatten_folders,
                             lang=lang,
                             subtitles=subtitles,
                             sub_use_sr_metadata=sub_use_sr_metadata,
                             anime=anime,
                             dvd_order=dvd_order,
                             search_format=search_format,
                             paused=paused,
                             blacklist=blacklist,
                             whitelist=whitelist,
                             default_status_after=default_status_after,
                             scene=scene,
                             skip_downloaded=skip_downloaded))

    def remove_show(self, series_id, series_provider_id, full=False):
        show_obj = find_show(series_id, series_provider_id)

        if not show_obj:
            raise CantRemoveShowException('Failed removing show: Show does not exist')
        elif not hasattr(show_obj, 'series_id'):
            raise CantRemoveShowException('Failed removing show: Show does not have an series_provider_id id')
        elif self._is_being(show_obj.series_id, (ShowTaskActions.REMOVE,)):
            raise CantRemoveShowException("{} is already queued to be removed".format(show_obj))

        # remove other queued actions for this show.
        self.remove_task(series_id)

        self.put(ShowTaskForceRemove(series_id=series_id, series_provider_id=series_provider_id, full=full))


class ShowTaskActions(Enum):
    REFRESH = 'Refresh'
    ADD = 'Add'
    UPDATE = 'Update'
    FORCEUPDATE = 'Force Update'
    RENAME = 'Rename'
    SUBTITLE = 'Subtitle'
    REMOVE = 'Remove Show'


class ShowTask(Task):
    """
    Represents an item in the queue waiting to be executed

    Can be either:
    - show being added (may or may not be associated with a show object)
    - show being refreshed
    - show being updated
    - show being force updated
    - show being subtitled
    """

    def __init__(self, series_id, series_provider_id, action):
        super(ShowTask, self).__init__(action.value, action)
        self.series_id = series_id
        self.series_provider_id = series_provider_id

    def is_in_queue(self):
        return self in sickrage.app.show_queue.queue

    @property
    def show_name(self):
        show_obj = find_show(self.series_id, self.series_provider_id)
        return show_obj.name if show_obj else str(self.series_id)

    @property
    def is_loading(self):
        return False

    def run(self):
        show_obj = find_show(self.series_id, self.series_provider_id)
        if show_obj:
            WebSocketMessage('SHOW_QUEUE_STATUS_UPDATED', {'seriesSlug': show_obj.slug, 'showQueueStatus': show_obj.show_queue_status}).push()
        else:
            WebSocketMessage('SHOW_QUEUE_STATUS_UPDATED',
                             {'seriesSlug': f'{self.series_id}-{self.series_provider_id.slug}', 'action': self.action.name}).push()

    def finish(self):
        show_obj = find_show(self.series_id, self.series_provider_id)
        if show_obj:
            WebSocketMessage('SHOW_QUEUE_STATUS_UPDATED', {'seriesSlug': show_obj.slug, 'showQueueStatus': show_obj.show_queue_status}).push()
        else:
            WebSocketMessage('SHOW_QUEUE_STATUS_UPDATED',
                             {'seriesSlug': f'{self.series_id}-{self.series_provider_id.slug}', 'action': self.action.name}).push()


class ShowTaskAdd(ShowTask):
    def __init__(self, series_provider_id, series_id, show_dir, default_status, quality, flatten_folders, lang, subtitles, sub_use_sr_metadata, anime,
                 dvd_order, search_format, paused, blacklist, whitelist, default_status_after, scene, skip_downloaded):
        super(ShowTaskAdd, self).__init__(series_id, series_provider_id, ShowTaskActions.ADD)

        self.show_dir = show_dir
        self.default_status = default_status
        self.quality = quality
        self.flatten_folders = flatten_folders
        self.lang = lang
        self.subtitles = subtitles
        self.sub_use_sr_metadata = sub_use_sr_metadata
        self.anime = anime
        self.search_format = search_format
        self.dvd_order = dvd_order
        self.paused = paused
        self.blacklist = blacklist
        self.whitelist = whitelist
        self.default_status_after = default_status_after
        self.scene = scene
        self.skip_downloaded = skip_downloaded
        self.priority = TaskPriority.HIGH

    @property
    def show_name(self):
        """
        Returns the show name if there is a show object created, if not returns
        the dir that the show is being added to.
        """

        show_obj = find_show(self.series_id, self.series_provider_id)
        return show_obj.name if show_obj else os.path.basename(self.show_dir)

    @property
    def is_loading(self):
        """
        Returns True if we've gotten far enough to have a show object, or False
        if we still only know the folder name.
        """
        if find_show(self.series_id, self.series_provider_id):
            return True

    def run(self):
        super(ShowTaskAdd, self).run()

        start_time = time.time()

        sickrage.app.log.info("Started adding show {} from show dir: {}".format(self.show_name, self.show_dir))

        s = sickrage.app.series_providers[self.series_provider_id].search(self.series_id,
                                                                          language=self.lang or sickrage.app.config.general.series_provider_default_language,
                                                                          enable_cache=False)

        if not s:
            sickrage.app.alerts.error(
                _("Unable to add show"),
                _("Unable to look up the show in {} on {} using ID {}, not using the NFO. Delete .nfo and try adding "
                  "manually again.").format(self.show_dir, sickrage.app.series_providers[self.series_provider_id].name, self.series_id)
            )

            if sickrage.app.config.trakt.enable:
                title = self.show_dir.split("/")[-1]

                data = {
                    'shows': [
                        {
                            'title': title,
                            'ids': {sickrage.app.series_providers[self.series_provider_id].trakt_id: self.series_id}
                        }
                    ]
                }

                TraktAPI()["sync/watchlist"].remove(data)

            return self._finish_early()

        # this usually only happens if they have an NFO in their show dir which gave us a series id that has no
        # proper english version of the show
        try:
            s.seriesname
        except AttributeError:
            sickrage.app.log.warning(
                f"Show in {self.show_dir} has no name on {sickrage.app.series_providers[self.series_provider_id].name}, "
                f"probably the wrong language used to search with")

            sickrage.app.alerts.error(_("Unable to add show"),
                                      _(f"Show in {self.show_dir} has no name on {sickrage.app.series_providers[self.series_provider_id].name}, "
                                        f"probably the wrong language. Delete .nfo and add manually in the correct language"))

            return self._finish_early()

        # if the show has no episodes/seasons
        if not len(s):
            sickrage.app.log.warning("Show " + str(s['seriesname']) + " is on " + str(
                sickrage.app.series_providers[self.series_provider_id].name) + "but contains no season/episode "
                                                                               "data.")
            sickrage.app.alerts.error(_("Unable to add show"),
                                      _("Show ") + str(s['seriesname']) + _(" is on ") + str(sickrage.app.series_providers[self.series_provider_id].name) + _(
                                          " but contains no season/episode data."))
            return self._finish_early()

        try:
            # add show to database
            show_obj = TVShow(self.series_id, self.series_provider_id, lang=self.lang, location=self.show_dir)

            # set up initial values
            show_obj.subtitles = self.subtitles if self.subtitles is not None else sickrage.app.config.subtitles.default
            show_obj.sub_use_sr_metadata = self.sub_use_sr_metadata if self.sub_use_sr_metadata is not None else False
            show_obj.quality = self.quality if self.quality is not None else sickrage.app.config.general.quality_default
            show_obj.flatten_folders = self.flatten_folders if self.flatten_folders is not None else sickrage.app.config.general.flatten_folders_default
            show_obj.scene = self.scene if self.scene is not None else sickrage.app.config.general.scene_default
            show_obj.anime = self.anime if self.anime is not None else sickrage.app.config.general.anime_default
            show_obj.dvd_order = self.dvd_order if self.dvd_order is not None else False
            show_obj.search_format = self.search_format if self.search_format is not None else sickrage.app.config.general.search_format_default
            show_obj.skip_downloaded = self.skip_downloaded if self.skip_downloaded is not None else sickrage.app.config.general.skip_downloaded_default
            show_obj.paused = self.paused if self.paused is not None else False

            # set up default new/missing episode status
            sickrage.app.log.info("Setting all current episodes to the specified default status: " + str(self.default_status))
            show_obj.default_ep_status = self.default_status

            # save to database
            show_obj.save()

            if show_obj.anime:
                if self.blacklist:
                    show_obj.release_groups.set_black_keywords(self.blacklist)
                if self.whitelist:
                    show_obj.release_groups.set_white_keywords(self.whitelist)
        except SeriesProviderException as e:
            sickrage.app.log.warning(
                _("Unable to add show due to an error with ") + sickrage.app.series_providers[self.series_provider_id].name + ": {}".format(e))
            sickrage.app.alerts.error(_("Unable to add show due to an error with ") + sickrage.app.series_providers[self.series_provider_id].name + "")
            return self._finish_early()
        except MultipleShowObjectsException:
            sickrage.app.log.warning(_("The show in ") + self.show_dir + _(" is already in your show list, skipping"))
            sickrage.app.alerts.error(_('Show skipped'),
                                      _("The show in ") + self.show_dir + _(" is already in your show list"))
            return self._finish_early()
        except Exception as e:
            sickrage.app.log.error(_("Error trying to add show: {}").format(e))
            sickrage.app.log.debug(traceback.format_exc())
            raise self._finish_early()

        try:
            sickrage.app.log.debug(_("Attempting to retrieve show info from IMDb"))
            show_obj.load_imdb_info()
        except Exception as e:
            sickrage.app.log.debug(_("Error loading IMDb info: {}").format(e))
            sickrage.app.log.debug(traceback.format_exc())

        try:
            show_obj.load_episodes_from_series_provider()
        except Exception as e:
            sickrage.app.log.debug(_("Error with ") + show_obj.series_provider.name + _(", not creating episode list: {}").format(e))
            sickrage.app.log.debug(traceback.format_exc())

        try:
            show_obj.load_episodes_from_dir()
        except Exception as e:
            sickrage.app.log.debug("Error searching dir for episodes: {}".format(e))
            sickrage.app.log.debug(traceback.format_exc())

        show_obj.write_metadata(force=True)
        show_obj.populate_cache()

        if sickrage.app.config.trakt.enable:
            # if there are specific episodes that need to be added by trakt
            sickrage.app.trakt_searcher.manage_new_show(show_obj)

            # add show to trakt.tv library
            if sickrage.app.config.trakt.sync:
                sickrage.app.trakt_searcher.add_show_to_trakt_library(show_obj)

            if sickrage.app.config.trakt.sync_watchlist:
                sickrage.app.log.info("update watchlist")
                sickrage.app.notification_providers['trakt'].update_watchlist(show_obj)

        # Retrieve scene exceptions
        show_obj.retrieve_scene_exceptions()

        # Load XEM data to DB for show
        xem_refresh(show_obj.series_id, show_obj.series_provider_id, force=True)

        # check if show has XEM mapping so we can determine if searches should go by scene numbering or series_provider_id
        # numbering.
        # if not self.scene and get_xem_numbering_for_show(show_obj.series_id, show_obj.series_provider_id):
        #     show_obj.scene = 1

        # if they set default ep status to WANTED then run the backlog to search for episodes
        if show_obj.default_ep_status == EpisodeStatus.WANTED:
            sickrage.app.log.info(_("Launching backlog for this show since it has episodes that are WANTED"))
            sickrage.app.backlog_searcher.search_backlog(show_obj.series_id, show_obj.series_provider_id)

        show_obj.default_ep_status = self.default_status_after

        show_obj.save()

        WebSocketMessage('SHOW_ADDED',
                         {'seriesSlug': show_obj.slug,
                          'series': show_obj.to_json(progress=True)}).push()

        sickrage.app.log.info("Finished adding show {} in {}s from show dir: {}".format(self.show_name, round(time.time() - start_time, 2), self.show_dir))

    def _finish_early(self):
        try:
            sickrage.app.show_queue.remove_show(self.series_id, self.series_provider_id)
        except CantRemoveShowException:
            WebSocketMessage('SHOW_REMOVED',
                             {'seriesSlug': f'{self.series_id}-{self.series_provider_id.slug}'}).push()


class ShowTaskRefresh(ShowTask):
    def __init__(self, series_id=None, series_provider_id=None, force=False):
        super(ShowTaskRefresh, self).__init__(series_id, series_provider_id, ShowTaskActions.REFRESH)

        # force refresh certain items
        self.force = force

    def run(self):
        super(ShowTaskRefresh, self).run()

        start_time = time.time()

        tv_show = find_show(self.series_id, self.series_provider_id)

        sickrage.app.log.info("Performing refresh for show: {}".format(tv_show.name))

        tv_show.refresh_dir()
        tv_show.write_metadata(force=self.force)
        tv_show.populate_cache(force=self.force)

        # Load XEM data to DB for show
        # xem_refresh(show.series_id, show.series_provider_id)

        tv_show.last_refresh = datetime.datetime.now()

        tv_show.save()

        WebSocketMessage('SHOW_REFRESHED',
                         {'seriesSlug': tv_show.slug,
                          'series': tv_show.to_json(episodes=True, details=True)}).push()

        sickrage.app.log.info("Finished refresh in {}s for show: {}".format(round(time.time() - start_time, 2), tv_show.name))


class ShowTaskRename(ShowTask):
    def __init__(self, series_id=None, series_provider_id=None):
        super(ShowTaskRename, self).__init__(series_id, series_provider_id, ShowTaskActions.RENAME)

    def run(self):
        super(ShowTaskRename, self).run()

        tv_show = find_show(self.series_id, self.series_provider_id)

        sickrage.app.log.info("Performing renames for show: {}".format(tv_show.name))

        if not os.path.isdir(tv_show.location):
            sickrage.app.log.warning("Can't perform rename on " + tv_show.name + " when the show dir is missing.")
            return

        ep_obj_rename_list = []

        for cur_ep_obj in (x for x in tv_show.episodes if x.location):
            # Only want to rename if we have a location
            if cur_ep_obj.location:
                if cur_ep_obj.related_episodes:
                    # do we have one of multi-episodes in the rename list already
                    have_already = False
                    for cur_related_ep in cur_ep_obj.related_episodes + [cur_ep_obj]:
                        if cur_related_ep in ep_obj_rename_list:
                            have_already = True
                            break
                    if not have_already:
                        ep_obj_rename_list.append(cur_ep_obj)
                else:
                    ep_obj_rename_list.append(cur_ep_obj)

        for cur_ep_obj in ep_obj_rename_list:
            cur_ep_obj.rename()

        WebSocketMessage('SHOW_RENAMED',
                         {'seriesSlug': tv_show.slug,
                          'series': tv_show.to_json(episodes=True, details=True)}).push()

        sickrage.app.log.info("Finished renames for show: {}".format(tv_show.name))


class ShowTaskSubtitle(ShowTask):
    def __init__(self, series_id=None, series_provider_id=None):
        super(ShowTaskSubtitle, self).__init__(series_id, series_provider_id, ShowTaskActions.SUBTITLE)

    def run(self):
        super(ShowTaskSubtitle, self).run()

        tv_show = find_show(self.series_id, self.series_provider_id)

        sickrage.app.log.info("Started downloading subtitles for show: {}".format(tv_show.name))

        tv_show.download_subtitles()

        WebSocketMessage('SHOW_SUBTITLED',
                         {'seriesSlug': tv_show.slug,
                          'series': tv_show.to_json(episodes=True, details=True)}).push()

        sickrage.app.log.info("Finished downloading subtitles for show: {}".format(tv_show.name))


class ShowTaskUpdate(ShowTask):
    def __init__(self, series_id=None, series_provider_id=None, force=False, action=ShowTaskActions.UPDATE):
        super(ShowTaskUpdate, self).__init__(series_id, series_provider_id, action if not force else ShowTaskActions.FORCEUPDATE)
        self.force = force

    def run(self):
        super(ShowTaskUpdate, self).run()

        show_obj = find_show(self.series_id, self.series_provider_id)

        start_time = time.time()

        sickrage.app.log.info("Performing updates for show: {}".format(show_obj.name))

        try:
            sickrage.app.log.debug("Retrieving show info from " + show_obj.series_provider.name + "")
            show_obj.load_from_series_provider(cache=False)
        except SeriesProviderAttributeNotFound as e:
            sickrage.app.log.warning("Data retrieved from " + show_obj.series_provider.name + " was incomplete, aborting: {}".format(e))
            return
        except SeriesProviderException as e:
            sickrage.app.log.warning("Unable to contact " + show_obj.series_provider.name + ", aborting: {}".format(e))
            return

        try:
            sickrage.app.log.debug("Attempting to retrieve show info from IMDb")
            show_obj.load_imdb_info()
        except Exception as e:
            sickrage.app.log.warning("Error loading IMDb info for {}: {}".format(show_obj.series_provider.name, e))

        # get episodes from database
        db_episodes = {}
        for data in show_obj.episodes:
            if data.season not in db_episodes:
                db_episodes[data.season] = {}
            db_episodes[data.season].update({data.episode: True})

        # get episodes from a series provider
        try:
            series_provider_episodes = show_obj.load_episodes_from_series_provider()
        except SeriesProviderException as e:
            sickrage.app.log.warning(
                "Unable to get info from " + show_obj.series_provider.name + ", the show info will not be refreshed: {}".format(e))
            series_provider_episodes = None

        if not series_provider_episodes:
            sickrage.app.log.warning("No data returned from " + show_obj.series_provider.name + ", unable to update this show")
        else:
            # for each ep we found on series_provider_id delete it from the DB list
            for curSeason in series_provider_episodes:
                for curEpisode in series_provider_episodes[curSeason]:
                    if curSeason in db_episodes and curEpisode in db_episodes[curSeason]:
                        del db_episodes[curSeason][curEpisode]

            # remaining episodes in the DB list are not on the series_provider_id, just delete them from the DB
            for curSeason in db_episodes:
                for curEpisode in db_episodes[curSeason]:
                    sickrage.app.log.info("Permanently deleting episode " + str(curSeason) + "x" + str(curEpisode) + " from the database")
                    try:
                        show_obj.delete_episode(curSeason, curEpisode)
                    except EpisodeDeletedException:
                        continue

        show_obj.retrieve_scene_exceptions()

        WebSocketMessage('SHOW_UPDATED',
                         {'seriesSlug': show_obj.slug,
                          'series': show_obj.to_json(episodes=True, details=True)}).push()

        sickrage.app.log.info("Finished updates in {}s for show: {}".format(round(time.time() - start_time, 2), show_obj.name))


class ShowTaskForceRemove(ShowTask):
    def __init__(self, series_id=None, series_provider_id=None, full=False):
        super(ShowTaskForceRemove, self).__init__(series_id, series_provider_id, ShowTaskActions.REMOVE)

        # lets make sure this happens before any other high priority actions
        self.priority = TaskPriority.EXTREME
        self.full = full

    @property
    def is_loading(self):
        """
        Returns false cause we are removing the show.
        """
        return False

    def run(self):
        super(ShowTaskForceRemove, self).run()

        show_obj = find_show(self.series_id, self.series_provider_id)

        sickrage.app.log.info("Removing show: {}".format(show_obj.name))

        show_obj.delete_show(full=self.full)

        if sickrage.app.config.trakt.enable:
            try:
                sickrage.app.trakt_searcher.remove_show_from_trakt_library(show_obj)
            except Exception as e:
                sickrage.app.log.warning("Unable to delete show from Trakt: %s. Error: %s" % (show_obj.name, e))

        WebSocketMessage('SHOW_REMOVED',
                         {'seriesSlug': show_obj.slug}).push()

        sickrage.app.log.info("Finished removing show: {}".format(show_obj.name))



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import json
from urllib.parse import urlencode

import sickrage
from sickrage.core.enums import SeriesProviderID
from sickrage.core.websession import WebSession
from sickrage.notification_providers import NotificationProvider


class EMBYNotification(NotificationProvider):
    def __init__(self):
        super(EMBYNotification, self).__init__()
        self.name = 'emby'

    def notify_snatch(self, ep_name):
        if sickrage.app.config.emby.notify_on_snatch:
            self._notify_emby(self.notifyStrings[self.NOTIFY_SNATCH] + ': ' + ep_name)

    def notify_download(self, ep_name):
        if sickrage.app.config.emby.notify_on_download:
            self._notify_emby(self.notifyStrings[self.NOTIFY_DOWNLOAD] + ': ' + ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if sickrage.app.config.emby.notify_on_subtitle_download:
            self._notify_emby(self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD] + ' ' + ep_name + ": " + lang)

    def notify_version_update(self, new_version="??"):
        if sickrage.app.config.emby.enable:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._notify_emby(title + " - " + update_text + new_version)

    def _notify_emby(self, message, host=None, emby_apikey=None):
        """Handles notifying Emby host via HTTP API

        Returns:
            Returns True for no issue or False if there was an error

        """

        # fill in omitted parameters
        if not host:
            host = sickrage.app.config.emby.host
        if not emby_apikey:
            emby_apikey = sickrage.app.config.emby.apikey

        url = 'http://%s/emby/Notifications/Admin' % (host)
        values = {'Name': 'SiCKRAGE', 'Description': message,
                  'ImageUrl': 'https://www.sickrage.ca/favicon.ico'}
        data = json.dumps(values)

        headers = {
            'X-MediaBrowser-Token': emby_apikey,
            'Content-Type': 'application/json'
        }

        try:
            resp = WebSession().get(url, data=data, headers=headers)
            sickrage.app.log.debug('EMBY: HTTP response: {}'.format(resp.text.replace('\n', '')))
        except Exception as e:
            sickrage.app.log.warning('EMBY: Warning: Couldn\'t contact Emby at {}: {}'.format(url, e))
            return False

        return True

    def test_notify(self, host, emby_apikey):
        return self._notify_emby('This is a test notification from SiCKRAGE', host, emby_apikey)

    def mass_notify_login(self, ipaddress=""):
        if sickrage.app.config.emby.enable:
            update_text = self.notifyStrings[self.NOTIFY_LOGIN_TEXT]
            title = self.notifyStrings[self.NOTIFY_LOGIN]
            self._notify_emby(title + " - " + update_text.format(ipaddress))

    def update_library(self, show=None):
        """Handles updating the Emby Media Server host via HTTP API
        Returns: True for no issue or False if there was an error
        """

        if sickrage.app.config.emby.enable:
            if not sickrage.app.config.emby.host:
                sickrage.app.log.debug('EMBY: No host specified, check your settings')
                return False

            if show:
                if show.series_provider_id == SeriesProviderID.THETVDB:
                    provider = 'tvdb'
                else:
                    sickrage.app.log.warning('EMBY: Series provider unknown')
                    return False
                query = '?%sid=%s' % (provider, show.series_id)
            else:
                query = ''

            url = 'http://%s/emby/Library/Series/Updated%s' % (sickrage.app.config.emby.host, query)
            values = {}
            data = urlencode(values)

            headers = {
                'X-MediaBrowser-Token': sickrage.app.config.emby.apikey,
                'Content-Type': 'application/json'
            }

            try:
                resp = WebSession().get(url, data=data, headers=headers)
                sickrage.app.log.debug('EMBY: HTTP response: ' + resp.text.replace('\n', ''))
            except Exception as e:
                sickrage.app.log.warning('EMBY: Warning: Couldn\'t contact Emby at {}: {}'.format(url, e))
                return False

            return True



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################



import os

from sickrage.metadata_providers import MetadataProvider


class PS3Metadata(MetadataProvider):
    """
    Metadata generation class for Sony PS3.

    The following file structure is used:

    show_root/cover.jpg                         (poster)
    show_root/Season ##/filename.ext            (*)
    show_root/Season ##/filename.ext.cover.jpg  (episode thumb)
    """

    def __init__(self,
                 show_metadata=False,
                 episode_metadata=False,
                 fanart=False,
                 poster=False,
                 banner=False,
                 episode_thumbnails=False,
                 season_posters=False,
                 season_banners=False,
                 season_all_poster=False,
                 season_all_banner=False):

        MetadataProvider.__init__(self,
                                  show_metadata,
                                  episode_metadata,
                                  fanart,
                                  poster,
                                  banner,
                                  episode_thumbnails,
                                  season_posters,
                                  season_banners,
                                  season_all_poster,
                                  season_all_banner)

        self.name = "Sony PS3"

        self.poster_name = "cover.jpg"

        # web-ui metadata template
        self.eg_show_metadata = "<i>not supported</i>"
        self.eg_episode_metadata = "<i>not supported</i>"
        self.eg_fanart = "<i>not supported</i>"
        self.eg_poster = "cover.jpg"
        self.eg_banner = "<i>not supported</i>"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>.ext.cover.jpg"
        self.eg_season_posters = "<i>not supported</i>"
        self.eg_season_banners = "<i>not supported</i>"
        self.eg_season_all_poster = "<i>not supported</i>"
        self.eg_season_all_banner = "<i>not supported</i>"

    # Override with empty methods for unsupported features
    def retrieve_show_metadata(self, folder):
        # no show metadata generated, we abort this lookup function
        return None, None, None

    def create_show_metadata(self, show_obj, force=False):
        pass

    def get_show_file_path(self, show_obj):
        pass

    def create_episode_metadata(self, ep_obj, force=False):
        pass

    def create_fanart(self, show_obj, which=0, force=False):
        pass

    def create_banner(self, show_obj, which=0, force=False):
        pass

    def create_season_posters(self, show_obj, force=False):
        pass

    def create_season_banners(self, ep_obj):
        pass

    def create_season_all_poster(self, show_obj, force=False):
        pass

    def create_season_all_banner(self, show_obj, force=False):
        pass

    @staticmethod
    def get_episode_thumb_path(ep_obj):
        """
        Returns the path where the episode thumbnail should be stored. Defaults to
        the same path as the episode file but with a .cover.jpg extension.

        ep_obj: a TVEpisode instance for which to create the thumbnail
        """
        if os.path.isfile(ep_obj.location):
            tbn_filename = ep_obj.location + ".cover.jpg"
        else:
            return None

        return tbn_filename



# coding=utf-8
# Author: Bart Sommer <bart.sommer88@gmail.com>
#
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE. If not, see <http://www.gnu.org/licenses/>.


import re

from requests.utils import dict_from_cookiejar

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.exceptions import AuthException
from sickrage.core.helpers import bs4_parser, try_int, convert_size
from sickrage.search_providers import TorrentProvider


class ImmortalseedProvider(TorrentProvider):
    def __init__(self):
        super(ImmortalseedProvider, self).__init__('Immortalseed', 'https://immortalseed.me', True)

        # URLs
        self._urls.update({
            'login': '{base_url}/takelogin.php'.format(**self._urls),
            'search': '{base_url}/browse.php'.format(**self._urls),
            'rss': '{base_url}/rss.php'.format(**self._urls),
        })

        # custom settings
        self.custom_settings = {
            'username': '',
            'password': '',
            'passkey': '',
            'freeleech': False,
            'minseed': 0,
            'minleech': 0
        }

        # Proper Strings
        self.proper_strings = ['PROPER', 'REPACK']

        # Cache
        self.cache = ImmortalseedCache(self, min_time=20)

    def _check_auth(self):
        if not self.custom_settings['username'] or not self.custom_settings['password']:
            raise AuthException("Your authentication credentials for " + self.name + " are missing, check your config.")

        return True

    def _check_auth_from_data(self, data):
        if not self.custom_settings['passkey']:
            sickrage.app.log.warning('Invalid passkey. Check your settings')

        return True

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {
            'username': self.custom_settings['username'],
            'password': self.custom_settings['password'],
        }

        try:
            response = self.session.post(self.urls['login'], data=login_params).text
        except Exception:
            sickrage.app.log.warning("Unable to connect to provider")
            return False

        if re.search('Username or password incorrect!', response):
            sickrage.app.log.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []
        if not self.login():
            return results

        # Search Params
        search_params = {
            'do': 'search',
            'include_dead_torrents': 'no',
            'search_type': 't_name',
            'category': 0,
            'keywords': ''
        }

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {0}".format(mode))
            for search_string in search_strings[mode]:
                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: {0}".format(search_string))
                    search_params['keywords'] = search_string

                resp = self.session.get(self.urls['search'], params=search_params)
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        def process_column_header(td):
            td_title = ''
            if td.img:
                td_title = td.img.get('title', td.get_text(strip=True))
            if not td_title:
                td_title = td.get_text(strip=True)
            return td_title

        with bs4_parser(data) as html:
            torrent_table = html.find('table', id='sortabletable')
            torrent_rows = torrent_table('tr') if torrent_table else []

            # Continue only if at least one Release is found
            if len(torrent_rows) < 2:
                sickrage.app.log.debug("Data returned from provider does not contain any torrents")
                return results

            labels = [process_column_header(label) for label in torrent_rows[0]('td')]

            # Skip column headers
            for result in torrent_rows[1:]:
                try:
                    tooltip = result.find('div', class_='tooltip-target')
                    if not tooltip:
                        continue

                    title = tooltip.get_text(strip=True)

                    # skip if torrent has been nuked due to poor quality
                    if title.startswith('Nuked.'):
                        continue

                    download_url = result.find('img', title='Click to Download this Torrent in SSL!').parent['href']
                    if not all([title, download_url]):
                        continue

                    cells = result('td')
                    seeders = try_int(cells[labels.index('Seeders')].get_text(strip=True))
                    leechers = try_int(cells[labels.index('Leechers')].get_text(strip=True))
                    torrent_size = cells[labels.index('Size')].get_text(strip=True)
                    size = convert_size(torrent_size, -1)

                    results += [
                        {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                    ]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error("Failed parsing provider")

        return results


class ImmortalseedCache(TVCache):
    def _get_rss_data(self):
        params = {
            'secret_key': self.provider.custom_settings['passkey'],
            'feedtype': 'downloadssl',
            'timezone': '-5',
            'categories': '44,32,7,47,8,48,9',
            'showrows': '50',
        }

        return self.get_rss_feed(self.provider.urls['rss'], params=params)

    def _check_auth(self, data):
        return self.provider._check_auth_from_data(data)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import operator
import re
import threading
import time
import traceback

from sqlalchemy import orm

import sickrage
from sickrage.core.common import Quality, EpisodeStatus
from sickrage.core.databases.main import MainDB
from sickrage.core.exceptions import AuthException
from sickrage.core.helpers import remove_non_release_groups, flatten
from sickrage.core.nameparser import InvalidNameException, InvalidShowException, NameParser
from sickrage.core.search import pick_best_result, snatch_episode
from sickrage.core.tv.show.helpers import find_show, get_show_list
from sickrage.search_providers import NZBProvider, NewznabProvider, TorrentProvider, TorrentRssProvider


class ProperSearcher(object):
    def __init__(self, *args, **kwargs):
        self.name = "PROPERSEARCHER"
        self.running = False

    def task(self, force=False):
        """
        Start looking for new propers
        :param force: Start even if already running (currently not used, defaults to False)
        """
        if self.running or not sickrage.app.config.general.download_propers:
            return

        try:
            self.running = True

            # set thread name
            threading.currentThread().setName(self.name)

            sickrage.app.log.info("Beginning the search for new propers")

            propers = self._get_proper_list()
            if propers:
                self._download_propers(propers)
            else:
                sickrage.app.log.info('No recently aired episodes, no propers to search for')

            sickrage.app.log.info("Completed the search for new propers")
        finally:
            self.running = False

    def _get_proper_list(self):
        """
        Walk providers for propers
        """

        session = sickrage.app.main_db.session()

        propers = {}
        final_propers = []

        search_date = datetime.datetime.today() - datetime.timedelta(days=2)

        orig_thread_name = threading.currentThread().getName()

        for show in get_show_list():
            wanted = self._get_wanted(show, search_date)
            if not wanted:
                sickrage.app.log.debug("Nothing needs to be downloaded for {}, skipping".format(show.name))
                continue

            self._lastProperSearch = self._get_last_proper_search(show.series_id, show.series_provider_id)

            # for each provider get a list of the
            for providerID, providerObj in sickrage.app.search_providers.sort(randomize=sickrage.app.config.general.randomize_providers).items():
                # check provider type and provider is enabled
                if not sickrage.app.config.general.use_nzbs and providerObj.provider_type in [NZBProvider.provider_type, NewznabProvider.provider_type]:
                    continue
                elif not sickrage.app.config.general.use_torrents and providerObj.provider_type in [TorrentProvider.provider_type, TorrentRssProvider.provider_type]:
                    continue
                elif not providerObj.is_enabled:
                    continue

                threading.currentThread().setName(orig_thread_name + " :: [" + providerObj.name + "]")

                sickrage.app.log.info("Searching for any new PROPER releases from " + providerObj.name)

                try:
                    for season, episode in wanted:
                        for x in providerObj.find_propers(show.series_id, show.series_provider_id, season, episode):
                            if not re.search(r'(^|[. _-])(proper|repack)([. _-]|$)', x.name, re.I):
                                sickrage.app.log.debug('Found a non-proper, we have caught and skipped it.')
                                continue

                            name = self._generic_name(x.name)
                            if name not in propers:
                                sickrage.app.log.debug("Found new proper: " + x.name)
                                x.provider = providerObj
                                propers[name] = x
                except AuthException as e:
                    sickrage.app.log.warning("Authentication error: {}".format(e))
                    continue
                except Exception as e:
                    sickrage.app.log.debug("Error while searching " + providerObj.name + ", skipping: {}".format(e))
                    sickrage.app.log.debug(traceback.format_exc())
                    continue

                threading.currentThread().setName(orig_thread_name)

            self._set_last_proper_search(show.series_id, show.series_provider_id, datetime.datetime.now())

        # take the list of unique propers and get it sorted by
        sorted_propers = sorted(propers.values(), key=operator.attrgetter('date'), reverse=True)
        for curProper in sorted_propers:
            try:
                parse_result = NameParser(False).parse(curProper.name)
            except InvalidNameException:
                sickrage.app.log.debug("Unable to parse the filename " + curProper.name + " into a valid episode")
                continue
            except InvalidShowException:
                sickrage.app.log.debug("Unable to parse the filename " + curProper.name + " into a valid show")
                continue

            if not parse_result.series_name:
                continue

            if not parse_result.episode_numbers:
                sickrage.app.log.debug("Ignoring " + curProper.name + " because it's for a full season rather than specific episode")
                continue

            show = find_show(parse_result.series_id, parse_result.series_provider_id)
            sickrage.app.log.debug("Successful match! Result " + parse_result.original_name + " matched to show " + show.name)

            # set the series_id in the db to the show's series_id
            curProper.series_id = parse_result.series_id

            # set the series_provider_id in the db to the show's series_provider_id
            curProper.series_provider_id = show.series_provider_id

            # populate our Proper instance
            curProper.season = parse_result.season_number if parse_result.season_number is not None else 1
            curProper.episode = parse_result.episode_numbers[0]
            curProper.release_group = parse_result.release_group
            curProper.version = parse_result.version
            curProper.quality = Quality.name_quality(curProper.name, parse_result.is_anime)
            curProper.content = None

            # filter release
            best_result = pick_best_result(curProper)
            if not best_result:
                sickrage.app.log.debug("Proper " + curProper.name + " were rejected by our release filters.")
                continue

            # only get anime proper if it has release group and version
            if show.is_anime:
                if not best_result.release_group and best_result.version == -1:
                    sickrage.app.log.debug("Proper " + best_result.name + " doesn't have a release group and version, ignoring it")
                    continue

            # check if we actually want this proper (if it's the right quality)            
            dbData = session.query(MainDB.TVEpisode).filter_by(series_id=best_result.series_id, season=best_result.season,
                                                               episode=best_result.episode).one_or_none()
            if not dbData:
                continue

            # only keep the proper if we have already retrieved the same quality ep (don't get better/worse ones)
            old_status, old_quality = Quality.split_composite_status(int(dbData.status))
            if old_status not in (EpisodeStatus.DOWNLOADED, EpisodeStatus.SNATCHED) or old_quality != best_result.quality:
                continue

            # check if we actually want this proper (if it's the right release group and a higher version)
            if show.is_anime:
                old_version = int(dbData.version)
                old_release_group = dbData.release_group
                if not -1 < old_version < best_result.version:
                    continue

                sickrage.app.log.info("Found new anime v" + str(best_result.version) + " to replace existing v" + str(old_version))

                if old_release_group != best_result.release_group:
                    sickrage.app.log.info("Skipping proper from release group: {}, does not match existing release "
                                          "group: {}".format(best_result.release_group, old_release_group))
                    continue

            # if the show is in our list and there hasn't been a proper already added for that particular episode
            # then add it to our list of propers
            if best_result.series_id != -1 and (best_result.series_id, best_result.season, best_result.episode) not in map(
                    operator.attrgetter('series_id', 'season', 'episode'), final_propers):
                sickrage.app.log.info("Found a proper that we need: " + str(best_result.name))
                final_propers.append(best_result)

        return final_propers

    def _get_wanted(self, show, search_date):
        session = sickrage.app.main_db.session()

        wanted = []

        for result in session.query(MainDB.TVEpisode).filter_by(series_id=show.series_id).filter(MainDB.TVEpisode.airdate >= search_date,
                                                                                                 MainDB.TVEpisode.status.in_(flatten(
                                                                                                         [EpisodeStatus.composites(EpisodeStatus.DOWNLOADED),
                                                                                                          EpisodeStatus.composites(EpisodeStatus.SNATCHED),
                                                                                                          EpisodeStatus.composites(
                                                                                                                  EpisodeStatus.SNATCHED_BEST)]))):
            wanted += [(result.season, result.episode)]

        return wanted

    def _download_propers(self, proper_list):
        """
        Download proper (snatch it)

        :param proper_list:
        """

        session = sickrage.app.main_db.session()

        for curProper in proper_list:
            history_limit = datetime.datetime.today() - datetime.timedelta(days=30)

            # make sure the episode has been downloaded before
            history_results = [x for x in
                               session.query(MainDB.History).filter_by(series_id=curProper.series_id, season=curProper.season, episode=curProper.episode,
                                                                       quality=curProper.quality).filter(MainDB.History.date >= history_limit,
                                                                                                         MainDB.History.action.in_(flatten(
                                                                                                             [EpisodeStatus.composites(EpisodeStatus.SNATCHED),
                                                                                                              EpisodeStatus.composites(
                                                                                                                  EpisodeStatus.DOWNLOADED)])))]

            # if we didn't download this episode in the first place we don't know what quality to use for the proper
            # so we can't do it
            if len(history_results) == 0:
                sickrage.app.log.info("Unable to find an original history entry for proper {} so I'm not downloading "
                                      "it.".format(curProper.name))
                continue

            # make sure that none of the existing history downloads are the same proper we're trying to download
            is_same = False
            clean_proper_name = self._generic_name(remove_non_release_groups(curProper.name))

            for curResult in history_results:
                # if the result exists in history already we need to skip it
                if self._generic_name(
                        remove_non_release_groups(curResult.resource)) == clean_proper_name:
                    is_same = True
                    break

            if is_same:
                sickrage.app.log.debug("This proper is already in history, skipping it")
                continue

            # make the result object
            result = curProper.provider.get_result(curProper.season, [curProper.episode])
            result.series_id = curProper.series_id
            result.url = curProper.url
            result.name = curProper.name
            result.quality = curProper.quality
            result.release_group = curProper.release_group
            result.version = curProper.version
            result.seeders = curProper.seeders
            result.leechers = curProper.leechers
            result.size = curProper.size
            result.files = curProper.files
            result.content = curProper.content

            # snatch it
            snatch_episode(result, EpisodeStatus.SNATCHED_PROPER)
            time.sleep(sickrage.app.config.general.cpu_preset.value)

    def _generic_name(self, name):
        return name.replace(".", " ").replace("-", " ").replace("_", " ").lower()

    def _set_last_proper_search(self, series_id, series_provider_id, when):
        """
        Record last propersearch in DB

        :param when: When was the last proper search
        """

        sickrage.app.log.debug("Setting the last proper search in database to " + str(when))

        try:
            show = find_show(series_id, series_provider_id)
            show.last_proper_search = when
            show.save()
        except orm.exc.NoResultFound:
            pass

    def _get_last_proper_search(self, series_id, series_provider_id):
        """
        Find last propersearch from DB
        """

        sickrage.app.log.debug("Retrieving the last check time from the DB")

        try:
            show = find_show(series_id, series_provider_id)
            return show.last_proper_search
        except orm.exc.NoResultFound:
            return 1



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import datetime

from dateutil.tz import gettz
from tornado.web import authenticated

import sickrage
from sickrage.core.helpers import try_int
from sickrage.core.tv.show.helpers import get_show_list
from sickrage.core.webserver.handlers.base import BaseHandler


class CalendarHandler(BaseHandler):
    def get(self, *args, **kwargs):
        if sickrage.app.config.general.calendar_unprotected:
            return self.calendar()
        else:
            return self.calendar_auth()

    @authenticated
    def calendar_auth(self):
        return self.calendar()

    def calendar(self):
        """ Provides a subscribeable URL for iCal subscriptions
        """

        utc = gettz('GMT')

        sickrage.app.log.info("Receiving iCal request from %s" % self.request.remote_ip)

        # Create a iCal string
        ical = 'BEGIN:VCALENDAR\r\n'
        ical += 'VERSION:2.0\r\n'
        ical += 'X-WR-CALNAME:SiCKRAGE\r\n'
        ical += 'X-WR-CALDESC:SiCKRAGE\r\n'
        ical += 'PRODID://SiCKRAGE Upcoming Episodes//\r\n'

        # Limit dates
        past_date = datetime.date.today() + datetime.timedelta(weeks=-52)
        future_date = datetime.date.today() + datetime.timedelta(weeks=52)

        # Get all the shows that are not paused and are currently on air (from kjoconnor Fork)
        for show in get_show_list():
            if show.status.lower() not in ['continuing', 'returning series'] or show.paused:
                continue

            for episode in show.episodes:
                if not past_date <= episode.airdate < future_date:
                    continue

                air_date_time = sickrage.app.tz_updater.parse_date_time(episode.airdate, show.airs,
                                                                        show.network).astimezone(utc)
                air_date_time_end = air_date_time + datetime.timedelta(minutes=try_int(show.runtime, 60))

                # Create event for episode
                ical += 'BEGIN:VEVENT\r\n'
                ical += 'DTSTART:' + air_date_time.strftime("%Y%m%d") + 'T' + air_date_time.strftime("%H%M%S") + 'Z\r\n'
                ical += 'DTEND:' + air_date_time_end.strftime("%Y%m%d") + 'T' + air_date_time_end.strftime(
                    "%H%M%S") + 'Z\r\n'
                if sickrage.app.config.general.calendar_icons:
                    ical += 'X-GOOGLE-CALENDAR-CONTENT-ICON:https://www.sickrage.ca/favicon.ico\r\n'
                    ical += 'X-GOOGLE-CALENDAR-CONTENT-DISPLAY:CHIP\r\n'
                ical += 'SUMMARY: {0} - {1}x{2} - {3}\r\n'.format(show.name, episode.season, episode.episode, episode.name)
                ical += 'UID:SiCKRAGE-' + str(datetime.date.today().isoformat()) + '-' + \
                        show.name.replace(" ", "-") + '-E' + str(episode.episode) + \
                        'S' + str(episode.season) + '\r\n'
                if episode.description:
                    ical += 'DESCRIPTION: {0} on {1} \\n\\n {2}\r\n'.format(
                        (show.airs or '(Unknown airs)'),
                        (show.network or 'Unknown network'),
                        episode.description.splitlines()[0])
                else:
                    ical += 'DESCRIPTION:' + (show.airs or '(Unknown airs)') + ' on ' + (
                            show.network or 'Unknown network') + '\r\n'

                ical += 'END:VEVENT\r\n'

        # Ending the iCal
        ical += 'END:VCALENDAR'

        return ical



# Author: Nick Sologoub
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import re
from urllib import parse

from requests.utils import dict_from_cookiejar

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, convert_size
from sickrage.search_providers import TorrentProvider


class PretomeProvider(TorrentProvider):
    def __init__(self):
        super(PretomeProvider, self).__init__("Pretome", 'https://pretome.info', True)

        self._urls.update({
            'login': '{base_url}/takelogin.php'.format(**self._urls),
            'detail': '{base_url}/details.php?id=%s'.format(**self._urls),
            'search': '{base_url}/browse.php?search=%s%s'.format(**self._urls),
            'download': '{base_url}/download.php/%s/%s.torrent'.format(**self._urls)
        })

        # custom settings
        self.custom_settings = {
            'username': '',
            'password': '',
            'pin': '',
            'minseed': 0,
            'minleech': 0
        }

        self.categories = "&st=1&cat%5B%5D=7"

        self.proper_strings = ['PROPER', 'REPACK']

        self.cache = TVCache(self, min_time=30)

    def _check_auth(self):

        if not self.custom_settings['username'] or not self.custom_settings['password'] or not self.custom_settings['pin']:
            sickrage.app.log.warning("Invalid username or password or pin. Check your settings")

        return True

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {'username': self.custom_settings['username'],
                        'password': self.custom_settings['password'],
                        'login_pin': self.custom_settings['pin']}

        try:
            response = self.session.post(self.urls['login'], data=login_params, timeout=30).text
        except Exception:
            sickrage.app.log.warning("Unable to connect to provider")
            return False

        if re.search('Username or password incorrect', response):
            sickrage.app.log.warning(
                "Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        if not self.login():
            return results

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: %s" % mode)
            for search_string in search_strings[mode]:
                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: %s " % search_string)

                search_url = self.urls['search'] % (parse.quote(search_string), self.categories)

                resp = self.session.get(search_url)
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        with bs4_parser(data) as html:
            # Continue only if one Release is found
            empty = html.find('h2', text="No .torrents fit this filter criteria")
            if empty:
                sickrage.app.log.debug("Data returned from provider does not contain any torrents")
                return results

            torrent_table = html.find('table', attrs={'style': 'border: none; width: 100%;'})
            if not torrent_table:
                sickrage.app.log.debug("Could not find table of torrents")
                return results

            torrent_rows = torrent_table.find_all('tr', attrs={'class': 'browse'})

            for result in torrent_rows:
                cells = result.find_all('td')
                size = None
                link = cells[1].find('a', attrs={'style': 'font-size: 1.25em; font-weight: bold;'})

                torrent_id = link['href'].replace('details.php?id=', '')

                try:
                    if 'title' in link:
                        title = link['title']
                    else:
                        title = link.contents[0]

                    download_url = self.urls['download'] % (torrent_id, link.contents[0])
                    seeders = int(cells[9].contents[0])
                    leechers = int(cells[10].contents[0])

                    # Need size for failed downloads handling
                    if size is None:
                        if re.match(r'[0-9]+,?\.?[0-9]*[KkMmGg]+[Bb]+', cells[7].text):
                            size = convert_size(cells[7].text, -1)

                    if not all([title, download_url]):
                        continue

                    results += [
                        {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                    ]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error("Failed parsing provider.")

        return results



from __future__ import absolute_import, division, print_function

from trakt.core.errors import ERRORS


class RequestFailedError(Exception):
    pass


class RequestError(Exception):
    def __init__(self, response):
        self.response = response
        self.status_code = response.status_code if response is not None else None

        self.error = ERRORS.get(self.status_code, ('Unknown', 'Unknown'))

        # Call super class with message
        super(RequestError, self).__init__('%s - "%s"' % self.error)


class ClientError(RequestError):
    pass


class ServerError(RequestError):
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import os

from tornado.escape import json_encode
from tornado.web import authenticated

from sickrage.core.helpers.browser import foldersAtPath
from sickrage.core.webserver.handlers.base import BaseHandler


class WebFileBrowserHandler(BaseHandler):
    def initialize(self):
        self.set_header('Content-Type', 'application/json')

    @authenticated
    def get(self, *args, **kwargs):
        path = self.get_argument('path', '')
        include_files = self.get_argument('includeFiles', None)
        file_types = self.get_argument('fileTypes', '')

        return json_encode(foldersAtPath(path, True, bool(int(include_files) if include_files else False), file_types.split(',')))


class WebFileBrowserCompleteHandler(BaseHandler):
    def initialize(self):
        self.set_header('Content-Type', 'application/json')

    @authenticated
    def get(self, *args, **kwargs):
        term = self.get_argument('term')
        include_files = self.get_argument('includeFiles', None)
        file_types = self.get_argument('fileTypes', '')

        return json_encode([entry['path'] for entry in foldersAtPath(
            os.path.dirname(term),
            includeFiles=bool(int(include_files) if include_files else False),
            fileTypes=file_types.split(',')
        ) if 'path' in entry])



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
from marshmallow_enum import EnumField
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema

from sickrage.core.common import Qualities, EpisodeStatus
from sickrage.core.databases.config import ConfigDB
from sickrage.core.enums import UserPermission, CheckPropersInterval, NzbMethod, ProcessMethod, FileTimestampTimezone, CpuPreset, MultiEpNaming, \
    DefaultHomePage, TorrentMethod, SearchFormat, PosterSortDirection, HomeLayout, PosterSortBy, HistoryLayout, TimezoneDisplay, UITheme, \
    TraktAddMethod, SeriesProviderID
from sickrage.core.helpers import camelcase
from sickrage.core.tv.show.coming_episodes import ComingEpsLayout, ComingEpsSortBy
from sickrage.notification_providers.nmjv2 import NMJv2Location
from sickrage.search_providers import SearchProviderType


class UsersSchema(SQLAlchemyAutoSchema):
    permissions = EnumField(UserPermission)

    class Meta:
        model = ConfigDB.Users
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class GeneralSchema(SQLAlchemyAutoSchema):
    proper_searcher_interval = EnumField(CheckPropersInterval)
    nzb_method = EnumField(NzbMethod)
    series_provider_default = EnumField(SeriesProviderID)
    process_method = EnumField(ProcessMethod)
    file_timestamp_timezone = EnumField(FileTimestampTimezone)
    cpu_preset = EnumField(CpuPreset)
    naming_multi_ep = EnumField(MultiEpNaming)
    naming_anime_multi_ep = EnumField(MultiEpNaming)
    default_page = EnumField(DefaultHomePage)
    status_default = EnumField(EpisodeStatus)
    status_default_after = EnumField(EpisodeStatus)
    ep_default_deleted_status = EnumField(EpisodeStatus)
    torrent_method = EnumField(TorrentMethod)
    quality_default = EnumField(Qualities)
    search_format_default = EnumField(SearchFormat)

    class Meta:
        model = ConfigDB.General
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class GUISchema(SQLAlchemyAutoSchema):
    poster_sort_dir = EnumField(PosterSortDirection)
    home_layout = EnumField(HomeLayout)
    coming_eps_layout = EnumField(ComingEpsLayout)
    coming_eps_sort = EnumField(ComingEpsSortBy)
    poster_sort_by = EnumField(PosterSortBy)
    history_layout = EnumField(HistoryLayout)
    timezone_display = EnumField(TimezoneDisplay)
    theme_name = EnumField(UITheme)

    class Meta:
        model = ConfigDB.GUI
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class BlackholeSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Blackhole
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SABnzbdSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.SABnzbd
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class NZBgetSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.NZBget
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SynologySchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Synology
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class TorrentSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Torrent
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class KodiSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Kodi
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class PlexSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Plex
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class EmbySchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Emby
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class GrowlSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Growl
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class FreeMobileSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.FreeMobile
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class TelegramSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Telegram
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class JoinSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Join
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class ProwlSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Prowl
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class TwitterSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Twitter
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class TwilioSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Twilio
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class Boxcar2Schema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Boxcar2
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class PushoverSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Pushover
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class LibnotifySchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Libnotify
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class NMJSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.NMJ
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class NMJv2Schema(SQLAlchemyAutoSchema):
    db_loc = EnumField(NMJv2Location)

    class Meta:
        model = ConfigDB.NMJv2
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SlackSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Slack
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class DiscordSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Discord
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class TraktSchema(SQLAlchemyAutoSchema):
    method_add = EnumField(TraktAddMethod)
    series_provider_default = EnumField(SeriesProviderID)

    class Meta:
        model = ConfigDB.Trakt
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class PyTivoSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.PyTivo
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class NMASchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.NMA
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class PushalotSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Pushalot
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class PushbulletSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Pushbullet
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class EmailSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Email
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class AlexaSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Alexa
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SubtitlesSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.Subtitles
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class FailedDownloadsSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.FailedDownloads
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class FailedSnatchesSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.FailedSnatches
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class AniDBSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.AniDB
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class QualitySizesSchema(SQLAlchemyAutoSchema):
    quality = EnumField(Qualities)

    class Meta:
        model = ConfigDB.QualitySizes
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SearchProvidersTorrentSchema(SQLAlchemyAutoSchema):
    provider_type = EnumField(SearchProviderType)

    class Meta:
        model = ConfigDB.SearchProvidersTorrent
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SearchProvidersNzbSchema(SQLAlchemyAutoSchema):
    provider_type = EnumField(SearchProviderType)

    class Meta:
        model = ConfigDB.SearchProvidersNzb
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SearchProvidersTorrentRssSchema(SQLAlchemyAutoSchema):
    provider_type = EnumField(SearchProviderType)

    class Meta:
        model = ConfigDB.SearchProvidersTorrentRss
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class SearchProvidersNewznabSchema(SQLAlchemyAutoSchema):
    provider_type = EnumField(SearchProviderType)

    class Meta:
        model = ConfigDB.SearchProvidersNewznab
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)


class MetadataProvidersSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = ConfigDB.MetadataProviders
        include_relationships = False
        load_instance = True

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = camelcase(field_obj.data_key or field_name)



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import datetime
import os

import sickrage
from sickrage.core import Quality
from sickrage.core.common import dateTimeFormat
from sickrage.core.helpers import convert_dict_keys_to_camelcase
from sickrage.core.tv.show.history import History
from sickrage.core.webserver.handlers.api.v2 import ApiV2BaseHandler


class ApiV2HistoryHandler(ApiV2BaseHandler):
    def get(self):
        """Get snatch and download history"
        ---
        tags: [History]
        summary: Get snatch and download history
        description: Get snatch and download history
        responses:
          200:
            description: Success payload
            content:
              application/json:
                schema:
                  HistorySuccessSchema
          400:
            description: Bad request; Check `errors` for any validation errors
            content:
              application/json:
                schema:
                  BadRequestSchema
          401:
            description: Returned if your JWT token is missing or expired
            content:
              application/json:
                schema:
                  NotAuthorizedSchema
          404:
            description: Returned if the given series slug does not exist or no series results.
            content:
              application/json:
                schema:
                  NotFoundSchema
        """

        limit = int(self.get_argument('limit', sickrage.app.config.gui.history_limit or 100))

        results = []

        for row in History().get(limit):
            status, quality = Quality.split_composite_status(int(row["action"]))

            # if self.type and not status.lower() == self.type:
            #     continue

            row["status"] = status.display_name
            row["quality"] = quality.name
            row["date"] = datetime.datetime.fromordinal(row["date"].toordinal()).timestamp()

            del row["action"]

            row["series_id"] = row.pop("series_id")
            row['series_provider_id'] = row['series_provider_id'].name
            row["resource_path"] = os.path.dirname(row["resource"])
            row["resource"] = os.path.basename(row["resource"])

            row = convert_dict_keys_to_camelcase(row)

            results.append(row)

        return self.json_response(results)
